from typing import Union

import numpy as np
import pyarrow as pa

import pandas as pd
import pandas.api.types as pdt
from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..base import BaseDataConverter


class PandasConverter(BaseDataConverter):
    dtype = "pd.DataFrame"
    supports_named_columns = True

    def to_numpy(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        return X.values

    def to_pandas(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        if isinstance(X, pd.Series):
            return X.to_frame(**get_kwargs(kwargs, X.to_frame))
        return X

    def to_series(self, X, **kwargs):
        if X.shape[1] == 1:
            return X.iloc[:, 0]
        if X.shape[0] == 1:
            return X.iloc[0, :]
        raise ValueError("Cannot convert multi-dimensional dataframe to series")

    def to_polars(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.from_pandas(X, **get_kwargs(kwargs, pl.from_pandas))

    def to_list(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        return X.values.tolist()

    def to_dict(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        return X.to_dict("list", **get_kwargs(kwargs, X.to_dict))

    def to_dicts(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        return X.to_dict("records", **get_kwargs(kwargs, X.to_dict))

    def to_arrow(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        if isinstance(X, pd.Series):
            return pa.Table.from_pandas(
                X.to_frame(), **pa_table_from_pandas_kwargs(kwargs)
            )
        return pa.Table.from_pandas(X, **pa_table_from_pandas_kwargs(kwargs))

    def to_dataset(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        if isinstance(X, pd.Series):
            X = X.to_frame(**get_kwargs(kwargs, X.to_frame))
        return Dataset.from_pandas(X, **get_kwargs(kwargs, Dataset.from_pandas))

    def to_bioset(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        if isinstance(X, pd.Series):
            X = X.to_frame(**get_kwargs(kwargs, X.to_frame))
        return Bioset.from_pandas(X, **get_kwargs(kwargs, Bioset.from_pandas))

    def to_iterabledataset(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for _, row in X.iterrows():
                yield row.to_dict()

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: Union[pd.DataFrame, pd.Series], **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_pandas(X, **get_kwargs(kwargs, dd.from_pandas))

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_pandas(X, **get_kwargs(kwargs, ray.data.from_pandas))

    def to_csr(self, X, **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(X)

    def append_column(self, X, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False)
        if isinstance(column, (pd.Series, pd.DataFrame)):
            if isinstance(column, pd.Series):
                column.name = name
            else:
                column.columns = [name]
            column.index = X.index
            return X.join(column)
        return pd.concat(
            X.join(pd.DataFrame({name: column}, index=X.index)), copy=False
        )

    def add_column(self, X, index, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False)
        if isinstance(column, (pd.Series, pd.DataFrame)):
            if isinstance(column, pd.Series):
                column.name = name
            else:
                column.columns = [name]
            column.index = X.index
            return X.join(column)
        if isinstance(X, pd.Series):
            X = X.to_frame()
        X.insert(index, name, column)
        return X

    def rename_column(self, X, name, new_name):
        name = self._check_column(X, name, str_only=False, as_list=False)
        new_name = self._check_column(X, new_name, str_only=False, as_list=False)
        return X.rename(columns={name: new_name})

    def rename_columns(self, X, mapping):
        mapping = {k: mapping.get(k, k) for k in X.columns}
        return X.rename(columns=mapping)

    def select_columns(
        self, X, columns=None, feature_type=None, **kwargs
    ) -> pd.DataFrame:
        if columns is None:
            return X
        if isinstance(columns, list) and len(columns) == 0:
            return pd.DataFrame()
        if isinstance(X, pd.Series):
            column = self._check_column(
                X, columns, str_only=True, as_list=False, convert=True
            )
            return X.to_frame(column)
        columns = self._check_column(
            X, columns, str_only=False, as_list=True, convert=False
        )
        if pdt.is_list_like(columns) and all(
            pdt.is_integer_dtype(type(c)) for c in columns
        ):
            return X.iloc[:, columns]
        return X.loc[:, columns]

    def select_column(self, X, column=None, **kwargs) -> pd.Series:
        if column is None:
            if isinstance(X, pd.DataFrame) and X.shape[1] == 1:
                return X.iloc[:, 0]
            return X
        column = self._check_column(
            X, column, str_only=False, as_list=False, convert=False
        )
        if isinstance(X, pd.Series):
            return (
                X if column is None or column == X.name or column == 0 else pd.Series()
            )
        if isinstance(column, int):
            return X.iloc[:, column]
        return X.loc[:, column]

    def set_column(self, X, column, value):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        X[column] = value
        return X

    def set_column_names(self, X, names, new_fingerprint=None):
        names = self._check_column(X, names, str_only=True, as_list=True, convert=False)
        if isinstance(X, pd.Series):
            X.name = names[0]
            return X
        X.columns = names
        return X

    def to_frame(self, X, name=None):
        if isinstance(X, pd.Series):
            name = self._check_column(
                X, name, str_only=True, as_list=False, convert=False
            )
            return X.to_frame(name)
        return X

    def select_rows(self, X, indices, **kwargs):
        return X.iloc[indices]

    def select_row(self, X, index):
        return X.iloc[index]

    def unique(self, X, column=None):
        return self.select_column(X, column).unique()

    def replace(self, X, column=None, mapping={}):
        return self.set_column(
            X, column, self.select_column(X, column).replace(mapping)
        )

    def nunique(self, X, column=None):
        return len(self.unique(X, column))

    def argmax(self, X, axis=0):
        return np.argmax(self.to_numpy(X), axis=axis)

    def ge(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) >= value
        return X >= value

    def gt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) > value
        return X > value

    def le(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) <= value
        return X <= value

    def lt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) < value
        return X < value

    def eq(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) == value
        return X == value

    def ne(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return self.select_column(X, column) != value
        return X != value

    def get_column_names(self, X, generate_cols=False):
        if isinstance(X, pd.Series):
            return [X.name]
        return X.columns.tolist()

    def get_shape(self, X):
        return X.shape

    def iter(self, X, batch_size, drop_last_batch=False):
        total_length = self.get_shape(X)[0]
        num_batches = total_length // batch_size
        if not drop_last_batch or total_length % batch_size == 0:
            num_batches += 1

        for batch_num in range(num_batches):
            start_idx = batch_num * batch_size
            end_idx = start_idx + batch_size
            yield X.iloc[start_idx:end_idx]

    def concat(self, tables, axis=0, **kwargs):
        if "ignore_index" not in kwargs:
            kwargs["ignore_index"] = True
            for i in range(len(tables)):
                if isinstance(tables[i], (pd.Series, pd.DataFrame)):
                    tables[i] = tables[i].reset_index(drop=True)
                if isinstance(tables[i], pd.Series):
                    tables[i] = tables[i].to_frame()
        if "copy" not in kwargs:
            kwargs["copy"] = False

        out = pd.concat(tables, axis=axis, **get_kwargs(kwargs, pd.concat))
        out.columns = [str(c) for tbl in tables for c in tbl.columns]
        return out

    def is_array_like(self, X):
        return True

    def get_dtypes(self, X, columns=None):
        dtype_map = {
            "object": "string",
            "double": "float64",
            "float": "float32",
            "int": "int64",
        }

        if columns is None:
            columns = self.get_column_names(X)
        dtypes = self.select_columns(X, columns).dtypes
        if isinstance(dtypes, pd.Series):
            return {
                k: dtype_map.get(str(v), str(v)) for k, v in dtypes.to_dict().items()
            }
        if isinstance(columns, list):
            columns = columns[0]
        return {columns: dtype_map.get(str(dtypes), str(dtypes))}

    def is_categorical(self, X, column, threshold=None):
        ser = self.select_column(X, column)

        if isinstance(ser.dtypes, pd.CategoricalDtype) or pd.api.types.is_string_dtype(
            ser.dtype
        ):
            return True
        if threshold and pd.api.types.is_integer_dtype(ser.dtype):
            return ser.nunique() / ser.shape[0] < threshold
        return False



from pathlib import Path
from typing import TYPE_CHECKING, Union

import numpy as np
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import get_kwargs

from ..arrow import ArrowConverter

if TYPE_CHECKING:
    import ray.data.dataset


class RayConverter(ArrowConverter):
    dtype = "Dataset"

    def to_numpy(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        return np.concatenate(
            [super().to_numpy(x, **kwargs) for x in X.to_arrow_refs()]
        )

    def to_pandas(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_pandas, "ray")
        import ray.data.dataset

        return X.to_pandas(**get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas))

    def to_polars(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_polars, "ray")
        requires_backends(self.to_polars, "polars")
        import polars as pl

        import ray.data.dataset

        return pl.from_pandas(
            X.to_pandas(**get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas)),
            **get_kwargs(kwargs, pl.from_pandas),
        )

    def to_arrow(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        return pa.concat_tables([x for x in X.to_arrow_refs()], promote="default")

    def to_list(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_list, "ray")
        import ray.data.dataset

        return X.to_pandas(
            **get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas)
        ).to_dict("list")

    def to_dict(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_dict, "ray")
        import ray.data.dataset

        return X.to_pandas(
            **get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas)
        ).to_dict()

    def to_dicts(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_dicts, "ray")
        import ray.data.dataset

        return X.to_pandas(
            **get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas)
        ).to_dict("records")

    def to_dataset(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset(self.to_arrow(X, kwargs), **get_kwargs(kwargs, Dataset.__init__))

    def to_bioset(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        return Bioset(self.to_arrow(X, kwargs), **get_kwargs(kwargs, Bioset.__init__))

    def to_iterabledataset(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_dicts, ["ray", "datasets"])
        from datasets import IterableDataset

        import ray.data.dataset

        def gen():
            for row in X.iter_rows(
                **get_kwargs(kwargs, ray.data.dataset.Dataset.iter_rows)
            ):
                yield row

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_file(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        path: Union[Path, str],
        **kwargs,
    ):
        path = Path(path)
        ext = Path(path).suffix
        valid_ext = [".arrow", ".parquet", ".json", ".csv"]
        if ext not in valid_ext:
            raise ValueError(
                f"Invalid file extension {ext}. Must be one of {valid_ext}"
            )

        if ext == ".arrow":
            path = path.replace(ext, "")
            path.mkdir(parents=True, exist_ok=True)
            X.write_parquet(path, **get_kwargs(kwargs))
        elif ext == ".parquet":
            path.parent.mkdir(parents=True, exist_ok=True)
            X.write_parquet(path, **get_kwargs(kwargs))
        elif ext == ".json":
            path.parent.mkdir(parents=True, exist_ok=True)
            X.write_json(path, **get_kwargs(kwargs))
        elif ext == ".csv":
            path.parent.mkdir(parents=True, exist_ok=True)
            X.write_csv(path, **get_kwargs(kwargs))

    def to_dask(
        self,
        X: Union["ray.data.dataset.MaterializedDataset", "ray.data.dataset.Dataset"],
        **kwargs,
    ):
        requires_backends(self.to_dask, "dask")
        requires_backends(self.to_dask, "ray")
        import dask.dataframe as dd

        import ray.data.dataset

        return dd.from_pandas(
            X.to_pandas(**get_kwargs(kwargs, ray.data.dataset.Dataset.to_pandas)),
            **get_kwargs(kwargs, dd.from_pandas),
        )

    def to_ray(self, X, **kwargs):
        return X



from typing import Dict

import numpy as np
import pandas as pd
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import get_kwargs, np_array_kwargs, np_asarray_kwargs

from ..arrow.arrow import ArrowConverter
from ..base import BaseDataConverter, get_data_format
from ..csr.csr import CSRConverter
from ..dask.dask import DaskConverter
from ..dataset.dataset import DatasetConverter
from ..iterabledataset.iterabledataset import IterableDatasetConverter
from ..list.list import ListConverter
from ..numpy.numpy import NumPyConverter
from ..pandas.pandas import PandasConverter
from ..polars.polars import PolarsConverter
from ..ray.ray import RayConverter

_FORMAT_TO_CONVERTER = {
    "np": NumPyConverter(),
    "numpy": NumPyConverter(),
    "list": ListConverter(),
    "pd": PandasConverter(),
    "pandas": PandasConverter(),
    "pl": PolarsConverter(),
    "polars": PolarsConverter(),
    "arrow": ArrowConverter(),
    "ds": DatasetConverter(),
    "dataset": DatasetConverter(),
    "ids": IterableDatasetConverter(),
    "iterabledataset": IterableDatasetConverter(),
    "iterable": IterableDatasetConverter(),
    "ray": RayConverter(),
    "sparse": CSRConverter(),
    "csr": CSRConverter(),
    "dask": DaskConverter(),
}


class DictConverter(BaseDataConverter):
    dtype = "dict"
    supports_named_columns = True

    def to_list(self, X: Dict[str, list], **kwargs):
        first_entry = next(iter(X.values()))
        ncol = len(first_entry)
        return [[X[k][i] for k in X] for i in range(ncol)]

    def to_dict(self, X: Dict[str, list], **kwargs):
        # This method is trivial for a DictConverter as X is already a dict
        return X

    def to_dicts(self, X: Dict[str, list], **kwargs):
        # Returns a list of dictionaries if the values are themselves dicts
        first_entry = next(iter(X.values()))
        return [{k: v[i] for k, v in X.items()} for i in range(len(first_entry))]

    def to_numpy(self, X: Dict[str, list], **kwargs):
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            return np.asarray(self.to_list(X, **kwargs), **np_asarray_kwargs(kwargs))
        return np.array(self.to_list(X, **kwargs), **np_array_kwargs(kwargs))

    def to_pandas(self, X: Dict[str, list], **kwargs):
        return pd.DataFrame(X, **get_kwargs(kwargs, pd.DataFrame.__init__))

    def to_polars(self, X: Dict[str, list], **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.DataFrame(X, **get_kwargs(kwargs, pl.DataFrame.__init__))

    def to_arrow(self, X: Dict[str, list], **kwargs):
        return pa.table(X)

    def to_dataset(self, X: Dict[str, list], **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset.from_dict(X, **get_kwargs(kwargs, Dataset.from_dict))

    def to_bioset(self, X: Dict[str, list], **kwargs):
        requires_backends(self.to_dataset, "biosets")
        from biosets import Bioset

        return Bioset.from_dict(X, **get_kwargs(kwargs, Bioset.from_dict))

    def to_iterabledataset(self, X: Dict[str, list], **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for key, value in X.items():
                yield {key: value}

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: Dict[str, list], **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_dict(X, **get_kwargs(kwargs, dd.from_dict))

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_arrow(
            pa.table(X), **get_kwargs(kwargs, ray.data.from_arrow)
        )

    def to_csr(self, X, **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(self.to_numpy(X), **get_kwargs(kwargs, csr_matrix.__init__))

    def append_column(self, X, name, column):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        X[name] = column
        return X

    def ge(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] >= value
        raise ValueError("Column must be specified for dict format")

    def gt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] > value
        raise ValueError("Column must be specified for dict format")

    def le(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] <= value
        raise ValueError("Column must be specified for dict format")

    def lt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] < value
        raise ValueError("Column must be specified for dict format")

    def eq(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] == value
        raise ValueError("Column must be specified for dict format")

    def ne(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X[column] != value
        raise ValueError("Column must be specified for dict format")

    def add_column(self, X, index, name, column):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        cols = list(X.keys())
        cols.insert(index, name)
        X[name] = column
        return {col: X[col] for col in cols}

    def rename_column(self, X, name, new_name):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        new_name = self._check_column(
            X, new_name, str_only=True, as_list=False, convert=False
        )
        X[new_name] = X.pop(name)
        return X

    def rename_columns(self, X, mapping):
        for name, new_name in mapping.items():
            X = self.rename_column(X, name, new_name)
        return X

    def select_columns(self, X, columns=None, **kwargs):
        if columns is None:
            return X
        if not columns:
            return dict()

        columns = self._check_column(
            X, columns, str_only=True, as_list=True, convert=True
        )
        return {k: X[k] for k in columns}

    def select_column(self, X, column, **kwargs):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=True
        )
        return X[column]

    def set_column(self, X, column, value):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        X[column] = value
        return X

    def set_column_names(self, X, names, new_fingerprint=None):
        names = self._check_column(X, names, str_only=True, as_list=True, convert=False)
        return {k: v for k, v in zip(names, X.values())}

    def to_frame(self, X, name=None):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=True)
        if name is not None:
            return {name: next(iter(X.values()))}
        return X

    def select_rows(self, X, indices, **kwargs):
        format = get_data_format(next(iter(X.values())))
        sel_rows = _FORMAT_TO_CONVERTER[format].select_rows
        return {
            k: sel_rows(v, indices, new_fingerprint=kwargs.get("new_fingerprint", None))
            for k, v in X.items()
        }

    def select_row(self, X, index):
        format = get_data_format(next(iter(X.values())))
        sel_row = _FORMAT_TO_CONVERTER[format].select_row

        return {k: sel_row(v, index) for k, v in X.items()}

    def unique(self, X, column=None):
        arr = self.select_column(X, column)
        return _FORMAT_TO_CONVERTER[get_data_format(arr)].unique(arr)

    def replace(self, X, column=None, mapping={}):
        arr = self.select_column(X, column)
        return _FORMAT_TO_CONVERTER[get_data_format(arr)].replace(arr)

    def nunique(self, X, column=None):
        return len(self.unique(X, column))

    def argmax(self, X, axis=0):
        format = get_data_format(next(iter(X.values())))
        tbl = _FORMAT_TO_CONVERTER[format].concat(list(X.values()), axis=0)
        return _FORMAT_TO_CONVERTER[format].argmax(tbl, axis=axis)

    def get_column_names(self, X, generate_cols=False):
        return list(X.keys())

    def get_shape(self, X):
        return len(next(iter(X.values()))), len(X)

    def iter(self, X, batch_size, drop_last_batch=False):
        total_length = self.get_shape(X)[0]
        # Calculate the number of batches
        num_batches = total_length // batch_size
        if not drop_last_batch or total_length % batch_size == 0:
            num_batches += 1
        format = get_data_format(next(iter(X.values())))
        sel_rows = _FORMAT_TO_CONVERTER[format].select_rows
        for batch_num in range(num_batches):
            start_idx = batch_num * batch_size
            end_idx = start_idx + batch_size
            yield {k: sel_rows(v, slice(start_idx, end_idx)) for k, v in X.items()}

    def concat(self, tables, axis=0, **kwargs):
        format = get_data_format(next(iter(tables[0].values())))
        concat = _FORMAT_TO_CONVERTER[format].concat
        if axis == 0:
            return {
                k: concat([table[k] for table in tables], axis=None) for k in tables[0]
            }
        out = tables[0]
        for t in tables[1:]:
            out.update(t)
        return out

    def is_array_like(self, X, min_row=10, min_col=10):
        if isinstance(next(iter(X.values())), (list, tuple, np.ndarray)):
            valid = True
            length = len(next(iter(X.values())))
            if length <= min_row or len(X) <= min_col:
                return False
            for v in X.values():
                if len(v) != length:
                    valid = False
                    break
            return valid

    def get_dtypes(self, X, columns=None):
        format = get_data_format(next(iter(X.values())))
        tbl = _FORMAT_TO_CONVERTER[format].concat(list(X.values()), axis=1)
        return _FORMAT_TO_CONVERTER[format].get_dtypes(tbl, columns=columns)

    def is_categorical(self, X, column, threshold=None):
        arr = self.select_column(X, column)
        format = get_data_format(arr)
        return _FORMAT_TO_CONVERTER[format].is_categorical(arr, threshold=threshold)



# ruff: noqa
from .import_util import (
    get_torch_version,
    get_transformers_version,
    is_accelerate_available,
    is_apex_available,
    is_auto_awq_available,
    is_auto_gptq_available,
    is_biosets_available,
    is_bitsandbytes_available,
    is_bs4_available,
    is_catboost_available,
    is_ccl_available,
    is_coloredlogs_available,
    is_cv2_available,
    is_cython_available,
    is_dask_available,
    is_dask_ml_available,
    is_datasets_available,
    is_decord_available,
    is_detectron2_available,
    is_essentia_available,
    is_faiss_available,
    is_flash_attn_2_available,
    is_flash_attn_available,
    is_flash_attn_greater_or_equal_2_10,
    is_flax_available,
    is_fsdp_available,
    is_ftfy_available,
    is_imblearn_available,
    is_in_notebook,
    is_ipex_available,
    is_ipywidgets_available,
    is_jieba_available,
    is_jinja_available,
    is_jumanpp_available,
    is_kenlm_available,
    is_keras_nlp_available,
    is_levenshtein_available,
    is_librosa_available,
    is_lightgbm_available,
    is_matplotlib_available,
    is_natten_available,
    is_ninja_available,
    is_nltk_available,
    is_onnx_available,
    is_openai_available,
    is_optimum_available,
    is_optimum_neuron_available,
    is_optuna_available,
    is_pandas_available,
    is_peft_available,
    is_phonemizer_available,
    is_plotly_available,
    is_polars_available,
    is_pretty_midi_available,
    is_protobuf_available,
    is_psutil_available,
    is_py3nvml_available,
    is_pyctcdecode_available,
    is_pypdf_available,
    is_pyspark_available,
    is_pytesseract_available,
    is_pytest_available,
    is_pytorch_quantization_available,
    is_ray_available,
    is_reportlab_available,
    is_rjieba_available,
    is_rpy2_arrow_available,
    is_rpy2_available,
    is_sacremoses_available,
    is_safetensors_available,
    is_sagemaker_dp_enabled,
    is_sagemaker_mp_enabled,
    is_scipy_available,
    is_seaborn_available,
    is_sentencepiece_available,
    is_seqio_available,
    is_sklearn_available,
    is_soundfile_availble,
    is_spacy_available,
    is_speech_available,
    is_statsmodels_available,
    is_sudachi_available,
    is_tensorflow_probability_available,
    is_tensorflow_text_available,
    is_tf2onnx_available,
    is_tf_available,
    is_timm_available,
    is_tokenizers_available,
    is_torch_available,
    is_torch_bf16_available,
    is_torch_bf16_available_on_device,
    is_torch_bf16_cpu_available,
    is_torch_bf16_gpu_available,
    is_torch_compile_available,
    is_torch_cuda_available,
    is_torch_fp16_available_on_device,
    is_torch_fx_available,
    is_torch_mps_available,
    is_torch_neuroncore_available,
    is_torch_npu_available,
    is_torch_sdpa_available,
    is_torch_tensorrt_fx_available,
    is_torch_tf32_available,
    is_torch_tpu_available,
    is_torch_xpu_available,
    is_torchaudio_available,
    is_torchdistx_available,
    is_torchdynamo_available,
    is_torchdynamo_compiling,
    is_torchvision_available,
    is_training_run_on_sagemaker,
    is_transformers_available,
    is_vision_available,
    is_xgboost_available,
    requires_backends,
    torch_only_method,
)
from .inspect import get_kwargs
from .py_util import (
    is_bioset,
    is_dataset,
    is_dataset_dict,
    is_iterable_dataset,
)



import numpy as np
import pandas as pd
import pandas.api.types as pdt
import pyarrow as pa

from biocore.utils import logging
from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    np_array_kwargs,
    np_asarray_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..base import BaseDataConverter

logger = logging.get_logger(__name__)


class ListConverter(BaseDataConverter):
    dtype = "list"

    def _is_fatten(self, X):
        return not pdt.is_list_like(X[0])

    def to_list(self, X, **kwargs):
        return X

    def to_dict(self, X, **kwargs):
        X = self.to_frame(X)

        names = kwargs.get("names", None)
        if names is None:
            if not isinstance(X[0], (list, tuple)):
                num_cols = 1
                X = [[x] for x in X]
            else:
                num_cols = len(X[0])
            names = [f"{i}" for i in range(num_cols)]
        return {name: [x[i] for x in X] for i, name in enumerate(names)}

    def to_dicts(self, X, **kwargs):
        X = self.to_frame(X)
        names = kwargs.get("names", None)
        if names is None:
            if not isinstance(X[0], (list, tuple)):
                pass
            else:
                len(X[0])
            names = [f"{i}" for i in range(len(X[0]))]
        return [dict(zip(names, x)) for x in X]

    def to_numpy(self, X, **kwargs):
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            return np.asarray(self.to_list(X, **kwargs), **np_asarray_kwargs(kwargs))
        return np.array(self.to_list(X, **kwargs), **np_array_kwargs(kwargs))

    def to_pandas(self, X, **kwargs):
        if self._is_fatten(X):
            return pd.Series(X, **get_kwargs(kwargs, pd.Series.__init__))
        return pd.DataFrame(X, **get_kwargs(kwargs, pd.DataFrame.__init__))

    def to_polars(self, X, **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        if self._is_fatten(X):
            return pl.Series(X, **get_kwargs(kwargs, pl.Series.__init__))
        return pl.from_records(X, **get_kwargs(kwargs, pl.from_records))

    def to_arrow(self, X, **kwargs):
        if self._is_fatten(X):
            return pa.array(X)
        return pa.Table.from_pandas(
            self.to_pandas(X, **kwargs), **pa_table_from_pandas_kwargs(kwargs)
        )

    def to_dataset(self, X, **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        input = self.to_pandas(X, **kwargs)
        if isinstance(input, pd.Series):
            input = input.to_frame()
        return Dataset.from_pandas(input, **get_kwargs(kwargs, Dataset.from_pandas))

    def to_bioset(self, X, **kwargs):
        requires_backends(self.to_dataset, "biosets")
        from biosets import Bioset

        input = self.to_pandas(X, **kwargs)
        if isinstance(input, pd.Series):
            input = input.to_frame()
        return Bioset.from_pandas(input, **get_kwargs(kwargs, Bioset.from_pandas))

    def to_iterabledataset(self, X, **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for x in X:
                yield x

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X, **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, dd.from_pandas)
        )

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_items(X, **get_kwargs(kwargs, ray.data.from_items))

    def to_csr(self, X, **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(self.to_numpy(X), **get_kwargs(kwargs, csr_matrix.__init__))

    def append_column(self, X, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        if self._is_fatten(X):
            X = np.asarray(X).reshape(-1, 1).tolist()
        if self._is_fatten(column):
            column = np.asarray(column).reshape(-1, 1).tolist()
        return np.concatenate((np.asarray(X), np.asarray(column)), axis=1).tolist()

    def add_column(self, X, index, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        if self._is_fatten(X):
            X = np.asarray(X).reshape(-1, 1).tolist()

        if not self._is_fatten(column):
            column = np.asarray(column).flatten()
        return np.insert(np.asarray(X), index, column, axis=1).tolist()

    def rename_column(self, X, name, new_name):
        logger.warning_once(
            "rename_column is not supported for list format. Returning the original data."
        )
        return X

    def rename_columns(self, X, mapping):
        logger.warning_once(
            "rename_columns is not supported for list format. Returning the original data."
        )
        return X

    def select_columns(self, X, columns=None, **kwargs):
        if columns is None:
            return X
        if not columns:
            return []

        columns = np.asarray(
            self._check_column(X, columns, str_only=False, as_list=True, convert=False)
        )
        if self._is_fatten(X):
            X = np.asarray(X).reshape(-1, 1).tolist()
        return np.asarray(X)[:, columns].tolist()

    def select_column(self, X, column, **kwargs):
        column = self._check_column(
            X, column, str_only=False, as_list=False, convert=False
        )
        if self._is_fatten(X) and column == 0:
            return X
        return np.asarray(X)[:, column].flatten().tolist()

    def set_column(self, X, column, values):
        column = self._check_column(
            X, column, str_only=False, as_list=False, convert=False
        )
        X = np.asarray(X)
        if self._is_fatten(X):
            return np.asarray(values).flatten().tolist()
        if not self._is_fatten(values):
            values = np.asarray(values).flatten()
        X[:, column] = values
        return X.tolist()

    def select_rows(self, X, indices, **kwargs):
        return [X[i] for i in indices]

    def select_row(self, X, index):
        return X[index]

    def unique(self, X, column=None):
        return list(set(self.select_column(X, column)))

    def replace(self, X, column=None, mapping={}):
        if not self._is_fatten(X):
            return self.set_column(
                X, column, [mapping.get(x[column], x[column]) for x in X]
            )
        return self.set_column(X, column, [mapping.get(x, x) for x in X])

    def argmax(self, X, axis=0):
        X = np.asarray(X)
        return np.argmax(X, axis=axis).tolist()

    def nunique(self, X, column=None):
        return len(self.unique(X, column))

    def get_column_names(self, X, generate_cols=False):
        if generate_cols:
            x_dims = self.get_shape(X)
            if len(x_dims) == 1:
                return [0]
            return list(range(x_dims[1]))
        return None

    def set_column_names(self, X, names, new_fingerprint=None):
        return X

    def get_shape(self, X):
        if self._is_fatten(X):
            return (len(X),)
        return (len(X), len(X[0]))

    def to_frame(self, X, name=None):
        if not self._is_fatten(X):
            return X
        return np.asarray(X).reshape(-1, 1).tolist()

    def iter(self, X, batch_size, drop_last_batch=False):
        total_length = self.get_shape(X)[0]
        # Calculate the number of batches
        num_batches = total_length // batch_size
        if not drop_last_batch or total_length % batch_size == 0:
            num_batches += 1

        for batch_num in range(num_batches):
            start_idx = batch_num * batch_size
            end_idx = start_idx + batch_size
            yield X[start_idx:end_idx]

    def concat(self, tables, axis=0, **kwargs):
        return np.concatenate(tables, axis=axis).tolist()

    def is_array_like(self, X, min_row=10, min_col=10):
        # check first 100 rows to see if all have the same length
        valid = True
        if self._is_fatten(X):
            return False
        length = len(X[0])
        if length <= min_col or len(X) <= min_row:
            return False
        for i in range(min(100, len(X))):
            if len(X[i]) != length:
                valid = False
                break
        return valid

    def get_dtypes(self, X, columns=None):
        full_name_mapper = {
            "str": "string",
            "int": "int64",
        }
        if self._is_fatten(X):
            return {0: full_name_mapper.get(type(X[0]).__name__, type(X[0]).__name__)}
        X = np.asarray(X)
        x_dims = X.shape
        dtype = X.dtype
        if pd.api.types.is_string_dtype(dtype):
            dtype = "string"
        if len(x_dims) == 1:
            return {0: str(dtype)}
        return {i: str(dtype) for i in range(x_dims[1])}

    def is_categorical(self, X, column, threshold=None):
        arr = self.select_column(X, column)
        if isinstance(arr[0], str):
            return True
        if threshold:
            n_unique = self.nunique(X, column)
            return n_unique / self.get_shape(X)[0] < threshold
        return False



import sys


def is_dataset_dict(val, iterable=True):
    if iterable:
        return isinstance(val, dict) and all(
            "datasets" in sys.modules
            and isinstance(
                v,
                (
                    sys.modules["datasets"].Dataset,
                    sys.modules["datasets"].IterableDataset,
                ),
            )
            for v in val.values()
        )

    return isinstance(val, dict) and all(
        "datasets" in sys.modules and isinstance(v, sys.modules["datasets"].Dataset)
        for v in val.values()
    )


def is_iterable_dataset(val):
    return "datasets" in sys.modules and isinstance(
        val, sys.modules["datasets"].IterableDataset
    )


def is_dataset(val, iterable=True):
    if iterable:
        return "datasets" in sys.modules and isinstance(
            val,
            (sys.modules["datasets"].Dataset, sys.modules["datasets"].IterableDataset),
        )

    return "datasets" in sys.modules and isinstance(
        val, sys.modules["datasets"].Dataset
    )


def is_bioset(val):
    return "biosets" in sys.modules and isinstance(val, sys.modules["biosets"].Bioset)



from pathlib import Path

import pandas as pd
import pyarrow as pa
import pyarrow.csv as pac
import pyarrow.feather as pf
import pyarrow.json as paj
import pyarrow.parquet as pq

from biocore.utils.import_util import is_datasets_available, requires_backends
from biocore.utils.inspect import get_kwargs

from ..base import BaseDataConverter


class IOConverter(BaseDataConverter):
    dtype = "io"

    def get_file_extension(self, path, **kwargs):
        return Path(path).suffix

    def validate_file(self, ext, valid_ext, **kwargs):
        if ext not in valid_ext:
            raise ValueError(
                f"Invalid file extension {ext}. Must be one of {valid_ext}"
            )

    def to_arrow(self, path, **kwargs):
        valid_ext = [".arrow", ".feather", ".parquet", ".json", ".csv"]
        ext = self.get_file_extension(path, **kwargs)
        self.validate_file(ext, valid_ext, **kwargs)
        if ext == ".arrow":
            if is_datasets_available():
                from datasets.table import InMemoryTable

                return InMemoryTable.from_file(path).table
            else:
                return pa.ipc.open_stream(path).read_all()

        elif ext == ".feather":
            return pf.read_table(path)
        elif ext == ".parquet":
            return pq.read_table(path)
        elif ext == ".json":
            return paj.read_json(path)
        elif ext == ".csv":
            return pac.read_csv(path)

    def to_pandas(self, path, **kwargs):
        return self.to_arrow(path, **kwargs).to_pandas()

    def to_polars(self, path, **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.from_arrow(self.to_arrow(path, **kwargs))

    def to_numpy(self, path, **kwargs):
        return self.to_pandas(path, **kwargs).values

    def to_list(self, path, **kwargs):
        return self.to_numpy(path, **kwargs).tolist()

    def to_dict(self, path, **kwargs):
        return self.to_pandas(path, **kwargs).to_dict(
            "list", **get_kwargs(kwargs, pd.DataFrame.to_dict)
        )

    def to_dicts(self, path, **kwargs):
        return self.to_pandas(path, **kwargs).to_dict(
            "records", **get_kwargs(kwargs, pd.DataFrame.to_dict)
        )

    def to_dataset(self, path, **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset, DatasetInfo, Features

        tbl = self.to_arrow(path, **kwargs)
        features = Features.from_arrow_schema(tbl.schema)
        info = DatasetInfo(features=features)
        return Dataset(tbl, info=info)

    def to_bioset(self, path, **kwargs):
        requires_backends(self.to_dataset, ["datasets", "biosets"])
        from biosets import Bioset
        from datasets import DatasetInfo, Features

        tbl = self.to_arrow(path, **kwargs)
        features = Features.from_arrow_schema(tbl.schema)
        info = DatasetInfo(features=features)
        return Bioset(tbl, info=info)

    def to_iterabledataset(self, path, **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        if isinstance(path, str):
            path = [path]

        def gen(**gen_kwargs):
            for f in path:
                yield self.to_dict(f, **gen_kwargs)

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_file(self, path, **kwargs):
        return path

    def to_dask(self, path, **kwargs):
        valid_ext = [".arrow", ".parquet", ".json", ".csv", ".hdf5", "h5"]
        ext = self.get_file_extension(path, **kwargs)
        self.validate_file(ext, valid_ext, **kwargs)
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        if ext == ".arrow":
            return dd.from_pandas(
                self.to_pandas(path, **kwargs), **get_kwargs(kwargs, dd.from_pandas)
            )
        elif path.endswith(".h5") or path.endswith(".hdf5"):
            return dd.read_hdf(path, **get_kwargs(kwargs, dd.read_hdf))
        elif ext == ".parquet":
            return dd.read_parquet(path, **get_kwargs(kwargs, dd.read_parquet))
        elif ext == ".json":
            return dd.read_json(path, **get_kwargs(kwargs, dd.read_json))
        elif ext == ".csv":
            return dd.read_csv(path, **get_kwargs(kwargs, dd.read_csv))



import sys
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import pyarrow as pa

import numpy as np
import pandas as pd
from biocore.utils.import_util import (
    is_biosets_available,
    is_dask_available,
    is_datasets_available,
    is_polars_available,
    requires_backends,
)
from biocore.utils.inspect import InvalidColumnSelectionError
from biocore.utils.py_util import is_dataset, is_iterable_dataset

from .arrow import ArrowConverter
from .base import BaseDataConverter, get_data_format
from .dask import DaskConverter
from .dataset import DatasetConverter
from .dict import DictConverter
from .dicts import DictsConverter
from .io import IOConverter
from .iterabledataset import IterableDatasetConverter
from .list import ListConverter
from .numpy import NumPyConverter
from .pandas import PandasConverter
from .polars import PolarsConverter
from .row_dict import RowDictConverter

if TYPE_CHECKING:
    from biosets import Bioset
    from datasets import Dataset, IterableDataset

    import polars as pl
    import ray.data.dataset

_FORMAT_TO_CONVERTER: Dict[str, BaseDataConverter] = {
    "np": NumPyConverter(),
    "numpy": NumPyConverter(),
    "list": ListConverter(),
    "dict": DictConverter(),
    "row_dict": RowDictConverter(),
    "dicts": DictsConverter(),
    "pd": PandasConverter(),
    "pandas": PandasConverter(),
    "arrow": ArrowConverter(),
    "io": IOConverter(),
    # "ray": RayConverter(),
    # "sparse": CSRConverter(),
    # "csr": CSRConverter(),
}

if is_dask_available():
    _FORMAT_TO_CONVERTER["dask"] = DaskConverter()
if is_polars_available():
    _FORMAT_TO_CONVERTER["pl"] = PolarsConverter()
    _FORMAT_TO_CONVERTER["polars"] = PolarsConverter()
if is_datasets_available():
    _FORMAT_TO_CONVERTER["dataset"] = DatasetConverter()
    _FORMAT_TO_CONVERTER["ds"] = DatasetConverter()
    _FORMAT_TO_CONVERTER["iterabledataset"] = IterableDatasetConverter()
    _FORMAT_TO_CONVERTER["iterable"] = IterableDatasetConverter()
    _FORMAT_TO_CONVERTER["ids"] = IterableDatasetConverter()
if is_biosets_available():
    _FORMAT_TO_CONVERTER["bioset"] = DatasetConverter()
    _FORMAT_TO_CONVERTER["bs"] = DatasetConverter()


SHORT_TO_LONG_FORMAT_NAMES = {
    "np": "numpy",
    "pd": "pandas",
    "pl": "polars",
    "ds": "dataset",
    "ids": "iterabledataset",
    "bs": "bioset",
}


class DataHandler:
    @staticmethod
    def get_format(X, options: Optional[Union[List[str], str]] = None):
        """Function to retrieve the source format and target format for the data.
        If source format is within the options, the target format is the same as the source format.
        Otherwise, the target format is the first option in the list.
        """
        if not options:
            from_format = get_data_format(X)
            return from_format

        from_format = get_data_format(X)

        to_format = options
        if isinstance(to_format, list):
            to_format = [SHORT_TO_LONG_FORMAT_NAMES.get(f, f) for f in to_format]
            if from_format in to_format:
                return from_format, from_format
            to_format = to_format[0]

        if to_format not in _FORMAT_TO_CONVERTER:
            raise ValueError(f"Unsupported format: {to_format}")

        return to_format

    @staticmethod
    def to_format(
        X,
        options: Optional[Union[List[str], str]] = None,
        input_columns=None,
        target_format=None,
        streaming=None,
        **kwargs,
    ):
        if X is None:
            raise ValueError("Cannot convert NoneType to a table format")
        from_format = get_data_format(X)
        if from_format is None:
            raise ValueError(f"Cannot determine format of {type(X)}")
        to_format = (
            SHORT_TO_LONG_FORMAT_NAMES.get(target_format, target_format)
            if target_format is not None
            else None
        )
        if not to_format:
            to_format = DataHandler.get_format(X, options=options)

        converter: BaseDataConverter = _FORMAT_TO_CONVERTER[from_format]

        if from_format == "iterabledataset" and streaming is not None:
            converter.streaming = streaming

        if input_columns:
            column_selector = DataHandler.select_columns
            if isinstance(input_columns, (int, str)):
                column_selector = DataHandler.select_column

            if DataHandler.supports_named_columns(from_format):
                return converter.converters()[to_format](
                    column_selector(X, input_columns), **kwargs
                )
            elif DataHandler.supports_named_columns(to_format):
                return column_selector(
                    converter.converters()[to_format](X, **kwargs), input_columns
                )
            else:
                return converter.converters()[to_format](
                    column_selector(X, input_columns), **kwargs
                )
        else:
            return converter.converters()[to_format](X, **kwargs)

    @staticmethod
    def to_pandas(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> pd.DataFrame:
        return DataHandler.to_format(
            X,
            target_format="pandas",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_numpy(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> np.ndarray:
        return DataHandler.to_format(
            X,
            target_format="numpy",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_polars(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> "pl.DataFrame":
        return DataHandler.to_format(
            X,
            target_format="polars",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_arrow(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> pa.Table:
        return DataHandler.to_format(
            X,
            target_format="arrow",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_dict(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> Dict:
        return DataHandler.to_format(
            X,
            target_format="dict",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_list(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> List:
        return DataHandler.to_format(
            X,
            target_format="list",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_dataset(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> "Dataset":
        return DataHandler.to_format(
            X,
            target_format="dataset",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_bioset(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> "Bioset":
        return DataHandler.to_format(
            X,
            target_format="bioset",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_iterabledataset(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> "IterableDataset":
        return DataHandler.to_format(
            X,
            target_format="iterabledataset",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_ray(
        X,
        input_columns=None,
        streaming=None,
        **kwargs,
    ) -> "ray.data.dataset.MaterializedDataset":
        return DataHandler.to_format(
            X,
            target_format="ray",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def to_csr(X, input_columns=None, streaming=None, **kwargs):
        return DataHandler.to_format(
            X,
            target_format="csr",
            input_columns=input_columns,
            streaming=streaming,
            **kwargs,
        )

    @staticmethod
    def supports_named_columns(data_or_format):
        if not isinstance(data_or_format, str):
            format = get_data_format(data_or_format)
        else:
            format = data_or_format
        converter = _FORMAT_TO_CONVERTER.get(format, None)
        if converter:
            return converter.supports_named_columns
        return False

    @staticmethod
    def append_column(X, name, column):
        format = get_data_format(X)
        col = DataHandler.to_format(column, target_format=format)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].append_column(X, name=name, column=col)

        raise ValueError(f"{type(X)} does not support column appending")

    @staticmethod
    def add_column(X, index, name, column):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].add_column(
                X, index=index, name=name, column=column
            )
        raise ValueError(f"{type(X)} does not support column adding")

    @staticmethod
    def select_rows(X, indices, **kwargs):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].select_rows(
                X, indices=indices, **kwargs
            )

        raise ValueError(f"{type(X)} does not support integer row selection")

    @staticmethod
    def select_row(X, index):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].select_row(X, index=index)

        raise ValueError(f"{type(X)} does not support integer row selection")

    @staticmethod
    def select_columns(X, columns=None, feature_type=None, **kwargs):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].select_columns(
                X, columns=columns, feature_type=feature_type, **kwargs
            )

        raise InvalidColumnSelectionError(
            f"{type(X)} does not support string column selection"
        )

    @staticmethod
    def drop_columns(X, columns):
        if columns is None:
            return X
        if isinstance(columns, str):
            columns = [columns]
        return DataHandler.select_columns(
            X, columns=DataHandler.exclude_column_names(X, columns)
        )

    @staticmethod
    def drop_column(X, column):
        if column is None:
            return X
        if isinstance(column, str):
            column = [column]
        return DataHandler.select_columns(
            X, columns=DataHandler.exclude_column_names(X, column)
        )

    @staticmethod
    def rename_column(X, name, new_name):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].rename_column(X, name, new_name)

        raise ValueError(f"{type(X)} does not support column renaming")

    @staticmethod
    def rename_columns(X, mapping):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].rename_columns(X, mapping=mapping)

        raise ValueError(f"{type(X)} does not support column renaming")

    @staticmethod
    def select_column(X, column, **kwargs):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].select_column(
                X, column=column, **kwargs
            )

    @staticmethod
    def unique(X, column=None):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].unique(X, column=column)

        raise ValueError(f"{type(X)} does not support unique")

    @staticmethod
    def replace(X, column=None, mapping={}):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].replace(
                X, column=column, mapping=mapping
            )

        raise ValueError(f"{type(X)} does not support replace")

    @staticmethod
    def nunique(X, column=None):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].nunique(X, column=column)

        raise ValueError(f"{type(X)} does not support nunique")

    @staticmethod
    def argmax(X, axis=0):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].argmax(X, axis=axis)

        raise ValueError(f"{type(X)} does not support argmax")

    @staticmethod
    def ge(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].ge(X, other)

        raise ValueError(f"{type(X)} does not support greater than or equal to")

    @staticmethod
    def gt(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].gt(X, other)

        raise ValueError(f"{type(X)} does not support greater than")

    @staticmethod
    def le(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].le(X, other)

        raise ValueError(f"{type(X)} does not support less than or equal to")

    @staticmethod
    def lt(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].lt(X, other)

        raise ValueError(f"{type(X)} does not support less than")

    @staticmethod
    def eq(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].eq(X, other)

        raise ValueError(f"{type(X)} does not support equal to")

    @staticmethod
    def ne(X, other):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].ne(X, other)

        raise ValueError(f"{type(X)} does not support not equal to")

    @staticmethod
    def set_column(X, column, value):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].set_column(
                X, column=column, value=value
            )

        raise ValueError(f"{type(X)} does not support setting columns")

    @staticmethod
    def set_column_names(X, names, new_fingerprint=None):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].set_column_names(
                X, names=names, new_fingerprint=new_fingerprint
            )

    @staticmethod
    def exclude_column_names(X, columns):
        if not columns:
            return X
        input_columns = DataHandler.get_column_names(X, generate_cols=True)

        input_map = dict.fromkeys(input_columns)
        if (
            isinstance(columns, list)
            and len(columns) > 0
            and isinstance(columns[0], int)
        ):
            columns = [input_columns[col] for col in columns]
        if isinstance(columns, str):
            columns = [columns]
        if isinstance(columns, int):
            columns = [input_columns[columns]]

        for col in columns:
            input_map.pop(col)
        return list(input_map.keys())

    @staticmethod
    def get_column_names(X, generate_cols=False):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].get_column_names(
                X, generate_cols=generate_cols
            )

    @staticmethod
    def get_column_names_by_feature_type(X, feature_type):
        requires_backends("get_column_names_by_feature_type", "datasets")
        from datasets import Dataset, IterableDataset

        if isinstance(X, (Dataset, IterableDataset)):
            if feature_type:
                return [
                    k
                    for k, v in X._info.features.items()
                    if isinstance(v, feature_type)
                ]
            else:
                return X.column_names
        raise ValueError(f"{type(X)} does not support feature type selection")

    @staticmethod
    def get_column_indices(X, col_names=None, raise_if_missing=True):
        if col_names is None:
            col_names = DataHandler.get_column_names(X, generate_cols=True)

        if len(col_names) > 0 and isinstance(col_names[0], int):
            return col_names
        col_map = {
            name: i
            for i, name in enumerate(
                DataHandler.get_column_names(X, generate_cols=True)
            )
        }

        return [
            col_map[name] for name in col_names if raise_if_missing or name in col_map
        ]

    @staticmethod
    def get_shape(X):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].get_shape(X)

        raise ValueError(f"{type(X)} does not support shape retrieval")

    @staticmethod
    def to_frame(X, name=None):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].to_frame(X, name=name)

        return X

    @staticmethod
    def iter(X, batch_size, drop_last_batch=False):
        if is_iterable_dataset(X):
            return X.iter(batch_size, drop_last_batch=drop_last_batch)
        if is_dataset(X):
            return X.iter(batch_size, drop_last_batch=drop_last_batch)

        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].iter(
                X, batch_size=batch_size, drop_last_batch=drop_last_batch
            )

        raise ValueError(f"{type(X)} does not support iteration")

    @staticmethod
    def concat(tables, axis=0, **kwargs):
        if len(tables) == 1:
            return tables[0]
        format = get_data_format(tables[0])
        tbls = [tables[0]]
        for table in tables[1:]:
            tbls.append(DataHandler.to_format(table, format, **kwargs))

        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].concat(tbls, axis=axis, **kwargs)

        raise ValueError(f"{type(tbls[0])} is not a recognized data type")

    @staticmethod
    def is_array_like(X, min_row=10, min_col=10):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].is_array_like(
                X, min_row=min_row, min_col=min_col
            )

        return False

    @staticmethod
    def get_dtypes(X, columns=None):
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].get_dtypes(X, columns=columns)

        raise ValueError(f"Cannot determine dtypes for {type(X)}")

    @staticmethod
    def get_numeric_features(
        X,
    ):
        if is_polars_available() and "polars" in sys.modules:
            import polars

        if isinstance(X, pd.DataFrame):
            return X.select_dtypes(include=["number"]).columns

        if isinstance(X, pa.Table):
            dtypes = {s.name: s.type for s in X.schema}
            return [
                k
                for k, dtype in dtypes.items()
                if pa.types.is_floating(dtype) or pa.types.is_integer(dtype)
            ]

        if is_polars_available() and isinstance(X, polars.DataFrame):
            import polars as pl
            import polars.selectors as plsel

            return X.select(plsel.by_dtype(pl.NUMERIC_DTYPES)).columns

        if isinstance(X, (Dataset, IterableDataset)):
            from biocore.features import NUMERIC_FEATURES

            dtypes = {s.name: s.type for s in X.data.table.schema}
            return [
                k
                for k, dtype in dtypes.items()
                if pa.types.is_floating(dtype)
                or pa.types.is_integer(dtype)
                or isinstance(X._info.features[k], NUMERIC_FEATURES)
            ]

        if isinstance(X, (list, dict, np.ndarray)):
            dtypes = DataHandler._check_first_n_rows(
                X,
                input_columns=DataHandler.get_column_names(X, generate_cols=True),
                n=1000,
            )
            return [
                k for k, dtype in dtypes.items() if pd.api.types.is_numeric_dtype(dtype)
            ]

        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].get_numeric_features(X)

        raise ValueError(f"Cannot determine if {type(X)} has numeric type features")

    @staticmethod
    def is_categorical(X, column, threshold=None):
        format = get_data_format(X)
        if threshold is not None and threshold > 1:
            threshold = min(threshold / DataHandler.get_shape(X)[0], 1)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].is_categorical(
                X, column=column, threshold=threshold
            )
        raise ValueError(f"Cannot determine if {type(X)} is categorical")

    @staticmethod
    def get_categorical_features(X):
        if isinstance(X, pd.DataFrame):
            return list(X.select_dtypes(include=["category", "object"]).columns)
        if isinstance(X, pd.Series):
            return list(
                X.to_frame().select_dtypes(include=["category", "object"]).columns
            )
        if isinstance(X, pa.Table):
            dtypes = {s.name: s.type for s in X.schema}
            return [
                k
                for k, dtype in dtypes.items()
                if pa.types.is_dictionary(dtype) or pa.types.is_string(dtype)
            ]

        if (
            is_polars_available()
            and "polars" in sys.modules
            and isinstance(X, sys.modules["polars"].DataFrame)
        ):
            import polars.selectors as plsel

            return X.select(plsel.by_dtype((pl.Categorical, pl.Object))).columns

        if is_datasets_available() and is_biosets_available():
            from datasets import Dataset, IterableDataset

            if isinstance(X, (Dataset, IterableDataset)):
                from biosets.features import CATEGORICAL_FEATURES

                return [
                    k
                    for k, v in X._info.features.items()
                    if pa.types.is_dictionary(v.pa_type)
                    or pa.types.is_string(v.pa_type)
                    or isinstance(v, CATEGORICAL_FEATURES)
                ]

        if isinstance(X, (list, dict)):
            dtypes = DataHandler._check_first_n_rows(
                X,
                input_columns=DataHandler.get_column_names(X, generate_cols=True),
                n=1000,
            )
            return [
                k
                for k, dtype in dtypes.items()
                if pd.api.types.is_object_dtype(dtype)
                or pd.api.types.is_string_dtype(dtype)
            ]
        format = get_data_format(X)
        if format in _FORMAT_TO_CONVERTER:
            return _FORMAT_TO_CONVERTER[format].get_categorical_features(X)

        raise ValueError(f"Cannot determine if {type(X)} is categorical")



import copy
from pathlib import Path
from typing import TYPE_CHECKING, Union

import pyarrow as pa

from biocore.utils.import_util import is_polars_available, requires_backends
from biocore.utils.inspect import get_kwargs, pa_table_from_pandas_kwargs
from biocore.utils.py_util import is_bioset

from ..arrow import ArrowConverter

if TYPE_CHECKING:
    from datasets import Dataset


class DatasetConverter(ArrowConverter):
    dtype = "Dataset"
    supports_named_columns = True

    def to_numpy(self, X: "Dataset", **kwargs):
        return super().to_numpy(self.to_arrow(X), **kwargs)

    def to_pandas(self, X: "Dataset", **kwargs):
        return X.to_pandas(**get_kwargs(kwargs, X.to_pandas))

    def to_polars(self, X: "Dataset", **kwargs):
        return X.to_polars(**get_kwargs(kwargs, X.to_polars))

    def to_arrow(self, X: "Dataset", **kwargs):
        if X._indices is not None:
            if is_polars_available():
                return self.to_polars(X, **kwargs).to_arrow()
            else:
                return pa.Table.from_pandas(
                    self.to_pandas(X, **kwargs), **pa_table_from_pandas_kwargs(kwargs)
                )
        return X.data.table

    def to_list(self, X: "Dataset", **kwargs):
        return super().to_list(self.to_arrow(X), **kwargs)

    def to_dict(self, X: "Dataset", **kwargs):
        return X.to_dict(**get_kwargs(kwargs, X.to_dict))

    def to_dicts(self, X: "Dataset", **kwargs):
        return super().to_dicts(self.to_arrow(X), **kwargs)

    def to_dataset(self, X: "Dataset", **kwargs):
        return X

    def to_bioset(self, X: "Dataset", **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        if isinstance(X, Bioset):
            return X

        return Bioset(
            X._data, info=X._info, fingerprint=X._fingerprint, indices_table=X._indices
        )

    def to_iterabledataset(self, X: "Dataset", **kwargs):
        return X.to_iterable_dataset(**get_kwargs(kwargs, X.to_iterable_dataset))

    def to_file(self, X: "Dataset", path: Union[Path, str], **kwargs):
        path = Path(path)
        ext = Path(path).suffix
        valid_ext = [".arrow", ".parquet", ".json", ".csv"]
        if ext not in valid_ext:
            raise ValueError(
                f"Invalid file extension {ext}. Must be one of {valid_ext}"
            )

        if ext == ".arrow":
            path = path.replace(ext, "")
            path.mkdir(parents=True, exist_ok=True)
            return X.save_to_disk(path, **get_kwargs(kwargs, X.save_to_disk))
        elif ext == ".parquet":
            path.parent.mkdir(parents=True, exist_ok=True)
            return X.to_parquet(path, **get_kwargs(kwargs, X.to_parquet))
        elif ext == ".json":
            path.parent.mkdir(parents=True, exist_ok=True)
            return X.to_json(path, **get_kwargs(kwargs, X.to_json))
        elif ext == ".csv":
            path.parent.mkdir(parents=True, exist_ok=True)
            return X.to_csv(path, **get_kwargs(kwargs, X.to_csv))

    def to_dask(self, X: "Dataset", **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_pandas(X.to_pandas(), **get_kwargs(kwargs, dd.from_pandas))

    def to_ray(self, X: "Dataset", **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_huggingface(
            X, **get_kwargs(kwargs, ray.data.from_huggingface)
        )

    def to_csr(self, X: "Dataset", **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(self.to_numpy(X), **get_kwargs(kwargs, csr_matrix.__init__))

    def append_column(self, X: "Dataset", name, column):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        return X.add_column(name, column=column, new_fingerprint=X._fingerprint)

    def add_column(self, X: "Dataset", index, name, column):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return self.append_column(X, name, column)

    def select_rows(self, X: "Dataset", indices, new_fingerprint=None):
        requires_backends(self.select_rows, "datasets")
        from datasets.table import Table

        if isinstance(indices, Table):
            X._indices = indices
            return X
        return X.select(
            indices=indices,
            new_fingerprint=new_fingerprint or X._fingerprint,
        )

    def select_row(self, X: "Dataset", index: "Dataset"):
        return X.select(index)

    def rename_column(self, X: "Dataset", name, new_name):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        new_name = self._check_column(
            X, new_name, str_only=False, as_list=False, convert=False
        )
        return X.rename_column(name, new_name)

    def rename_columns(self, X: "Dataset", mapping):
        mapping = {k: mapping.get(k, k) for k in X.column_names}
        return X.rename_columns(mapping)

    def select_columns(self, X: "Dataset", columns=None, feature_type=None, **kwargs):
        if columns:
            columns = self._check_column(
                X, columns, str_only=True, as_list=True, convert=True
            )
            new_fingerprint = kwargs.pop("new_fingerprint", None)
            keep_old_fingerprint = kwargs.pop("keep_old_fingerprint", True)
            if new_fingerprint is None and keep_old_fingerprint:
                new_fingerprint = X._fingerprint
            return X.select_columns(columns, new_fingerprint=new_fingerprint, **kwargs)
        elif feature_type:
            return X.select_columns(
                [k for k, v in X._info.features.items() if isinstance(v, feature_type)]
            )
        else:
            return X

    def select_column(self, X: "Dataset", column, **kwargs):
        return self.select_columns(X, column, **kwargs)

    def unique(self, X: "Dataset", column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=True
        )
        if column is None:
            if X.num_columns == 1:
                column = X.column_names[0]
            else:
                raise ValueError(
                    "Column must be specified for Dataset with multiple columns"
                )
        return X.unique(column)

    def replace(self, X: "Dataset", column=None, mapping={}):
        new_col = pa.array(
            [mapping.get(v.as_py(), v.as_py()) for v in self.to_arrow(X).column(column)]
        )
        return self.set_column(X, column, new_col)

    def nunique(self, X: "Dataset", column=None):
        return len(self.unique(X, column))

    def argmax(self, X: "Dataset", axis=0):
        return super().argmax(self.to_arrow(X), axis)

    def ge(self, X, value, column=None):
        return super().ge(self.to_arrow(X), value, column=column)

    def gt(self, X, value, column=None):
        return super().gt(self.to_arrow(X), value, column=column)

    def le(self, X, value, column=None):
        return super().le(self.to_arrow(X), value, column=column)

    def lt(self, X, value, column=None):
        return super().lt(self.to_arrow(X), value, column=column)

    def eq(self, X, value, column=None):
        return super().eq(self.to_arrow(X), value, column=column)

    def ne(self, X, value, column=None):
        return super().ne(self.to_arrow(X), value, column=column)

    def set_column(self, X: "Dataset", column, value):
        requires_backends(self.set_column, "datasets")

        from datasets.features.features import generate_from_arrow_type

        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        col_pos = X.column_names.index(column)
        out = copy.deepcopy(X)
        out._data.table = out._data.table.set_column(col_pos, column, value)
        feat = generate_from_arrow_type(out._data.table.column(col_pos).type)
        out._info.features[column] = feat
        return out

    def set_column_names(self, X: "Dataset", names, new_fingerprint=None):
        names = self._check_column(X, names, str_only=True, as_list=True, convert=False)
        return X.rename_columns(
            {old_name: new_name for old_name, new_name in zip(X.column_names, names)},
            new_fingerprint=new_fingerprint,
        )

    def get_column_names(self, X: "Dataset", generate_cols=False):
        return X.column_names

    def get_column_names_by_feature_type(self, X: "Dataset", feature_type):
        if feature_type:
            return [
                k for k, v in X._info.features.items() if isinstance(v, feature_type)
            ]
        return X.column_names

    def get_shape(self, X: "Dataset"):
        return X.shape

    def to_frame(self, X: "Dataset", name=None):
        return X

    def iter(self, X: "Dataset", batch_size, drop_last_batch=False):
        return X.iter(batch_size, drop_last_batch=drop_last_batch)

    def concat(self, tables, axis=0, **kwargs):
        if is_bioset(tables[0]):
            from biosets import concatenate_datasets

            return concatenate_datasets(tables, axis=axis)

        from datasets import concatenate_datasets

        return concatenate_datasets(tables, axis=axis)

    def is_array_like(self, X: "Dataset", min_row=10, min_col=10):
        return True

    def get_dtypes(self, X: "Dataset", columns=None):
        columns = self._check_column(X, columns)
        if columns is None:
            return {k: str(v.pa_type) for k, v in X._info.features.items()}
        return {col: X._info.features[col].dtype for col in columns}

    def is_categorical(self, X: "Dataset", column, threshold=None):
        requires_backends(self.is_categorical, ["datasets", "biosets"])
        from biosets.features import CATEGORICAL_FEATURES
        from datasets.features import Value

        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=True
        )
        feature = X._info.features[column]
        if isinstance(feature, CATEGORICAL_FEATURES):
            return True
        if isinstance(feature, Value):
            if feature.dtype == "string":
                return True
            if threshold and "int" in feature.dtype:
                n_unique = self.nunique(X, column)
                return n_unique / self.get_shape(X)[0] < threshold
            return False
        return False



import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, MutableMapping, Optional, Union

import pyarrow as pa

import biocore.utils.logging
import numpy as np
import pandas as pd
import pandas.api.types as pdt
from biocore.utils.import_util import (
    is_biosets_available,
    is_datasets_available,
    is_polars_available,
    is_ray_available,
)

if TYPE_CHECKING:
    pass

logger = biocore.utils.logging.get_logger(__name__)


def get_data_format(data):
    if is_ray_available() and "ray" in sys.modules:
        import ray.data.dataset

        if isinstance(data, ray.data.dataset.MaterializedDataset):
            return "ray"

    if isinstance(data, (Path, str)):
        return "io"
    if isinstance(data, np.ndarray):
        return "numpy"
    if isinstance(data, (pd.DataFrame, pd.Series)):
        return "pandas"
    if is_polars_available() and "polars" in sys.modules:
        import polars as pl

        if isinstance(data, (pl.DataFrame, pl.Series)):
            return "polars"
    if isinstance(data, (pa.Table, pa.Array, pa.ChunkedArray)):
        return "arrow"
    if is_biosets_available():
        from biosets import Bioset

        if isinstance(data, Bioset):
            return "bioset"
    if is_datasets_available():
        from datasets import Dataset, IterableDataset

        if isinstance(data, (Dataset, IterableDataset)):
            return "dataset" if isinstance(data, Dataset) else "iterabledataset"
    if isinstance(data, list):
        if len(data) == 0:
            return None
        if isinstance(data[0], dict):
            return "dicts"
        return "list"
    if isinstance(data, (dict, MutableMapping)):
        if len(data) == 0:
            return None
        first_entry = next(iter(data.values()))
        if isinstance(first_entry, list):
            return "dict"
        return "row_dict"

    return None


class BaseDataConverter:
    """
    A class that provides methods to convert data between different formats.
    """

    dtype = None
    supports_named_columns = False

    def converters(self):
        return {
            "np": self.to_numpy,
            "numpy": self.to_numpy,
            "list": self.to_list,
            "row_dict": self.to_dict,
            "dict": self.to_dict,
            "dicts": self.to_dicts,
            "pd": self.to_pandas,
            "pandas": self.to_pandas,
            "dataframe": self.to_pandas,
            "series": self.to_series,
            "pl": self.to_polars,
            "polars": self.to_polars,
            "arrow": self.to_arrow,
            "ds": self.to_dataset,
            "dataset": self.to_dataset,
            "ids": self.to_iterabledataset,
            "iterabledataset": self.to_iterabledataset,
            "iterable": self.to_iterabledataset,
            "bioset": self.to_bioset,
            "bs": self.to_bioset,
            "io": self.to_file,
            "dask": self.to_dask,
            "ray": self.to_ray,
            "sparse": self.to_csr,
            "csr": self.to_csr,
        }

    def to_list(self, X, **kwargs):
        raise NotImplementedError()

    def to_dict(self, X, **kwargs):
        raise NotImplementedError()

    def to_dicts(self, X, **kwargs):
        raise NotImplementedError()

    def to_numpy(self, X, **kwargs):
        raise NotImplementedError()

    def to_pandas(self, X, **kwargs):
        raise NotImplementedError()

    def to_series(self, X, **kwargs):
        raise NotImplementedError()

    def to_polars(self, X, **kwargs):
        raise NotImplementedError()

    def to_arrow(self, X, **kwargs):
        raise NotImplementedError()

    def to_dataset(self, X, **kwargs):
        raise NotImplementedError()

    def to_bioset(self, X, **kwargs):
        raise NotImplementedError()

    def to_iterabledataset(self, X, **kwargs):
        raise NotImplementedError()

    def to_file(self, X, **kwargs):
        raise NotImplementedError()

    def to_dask(self, X, **kwargs):
        raise NotImplementedError()

    def to_ray(self, X, **kwargs):
        raise NotImplementedError()

    def to_csr(self, X, **kwargs):
        raise NotImplementedError()

    def _check_first_n_rows(
        self,
        X: Union[List[List[Any]], List[Dict[str, List[Any]]], Dict[str, List[Any]]],
        input_columns: Optional[Union[List[str], List[int]]] = None,
        n: int = 100,
    ):
        if input_columns is None:
            cols = set(self.get_column_names(X, generate_cols=True))
        else:
            cols = set(input_columns)
        if isinstance(X, (list, np.ndarray)):
            if isinstance(X[0], (list, tuple, np.ndarray)):
                if not isinstance(next(iter(cols)), int):
                    if len(X[0]) != len(cols):
                        raise ValueError(
                            "Column indices must be integers when input is a list of lists or ndarrays"
                        )
                    cols = set(range(len(X[0])))
                dtypes = {
                    i: type(v)
                    for i, v in enumerate(X[0])
                    if i in cols and v is not None
                }
                null_dtypes = set(range(len(X[0]))) - set(dtypes.keys())
            elif isinstance(X[0], dict):
                dtypes = {
                    k: type(v) for k, v in X[0].items() if k in cols and v is not None
                }
                null_dtypes = set(X[0].keys()) - set(dtypes.keys())

            count = 0
            nrows = min(n, len(X))
            while len(null_dtypes) > 0 and count < nrows:
                count += 1
                to_rem = []
                for k in null_dtypes:
                    if X[count][k] is not None:
                        to_rem.append(k)
                        dtypes[k] = type(X[count][k])
                null_dtypes -= set(to_rem)
        elif isinstance(X, dict):
            if isinstance(X[next(iter(X))], (list, tuple, np.ndarray)):
                dtypes = {
                    k: type(v[0])
                    for k, v in X.items()
                    if k in cols and v[0] is not None
                }
                null_dtypes = set(X.keys()) - set(dtypes.keys())
                count = 0
                nrows = min(n, len(next(iter(X.values()))))
                while len(null_dtypes) > 0 and count < nrows:
                    to_rem = []
                    for k in null_dtypes:
                        if X[k][count] is not None:
                            to_rem.append(k)
                            dtypes[k] = type(X[k][count])
                    count += 1
                    null_dtypes -= set(to_rem)
            else:
                dtypes = {
                    k: type(v) for k, v in X.items() if k in cols and v is not None
                }
                null_dtypes = set(X.keys()) - set(dtypes.keys())
        else:
            raise ValueError(f"Cannot check first n rows for type {type(X)}")

        return dtypes

    def _check_column(
        self,
        X,
        columns: Optional[Union[str, List[str], int, List[int]]] = None,
        str_only=False,
        as_list=True,
        convert=True,
    ) -> Optional[Union[str, List[str], int, List[int]]]:
        """Check if the column is a string or integer and convert it to the appropriate format if necessary.

        Args:
            X (Any): the input data
            columns (Union[str, List[str], int, List[int]], *optional*): the column name or index
            str_only (bool, optional): whether to return only strings. Defaults to False.

        Raises:
            ValueError: _description_

        Returns:
            Union[str, List[str], int, List[int]]: the column name or index
        """
        if pdt.is_list_like(columns) or isinstance(columns, (slice, range)):
            if (
                columns
                and str_only
                and (
                    isinstance(columns, (slice, range))
                    or pdt.is_integer_dtype(type(columns[0]))
                )
            ):
                if convert:
                    cols = self.get_column_names(X)
                    if isinstance(columns, slice):
                        columns = cols[columns]
                    else:
                        columns = [cols[i] for i in columns]
                else:
                    columns = [f"__column_{i}__" for i in columns]
            if not as_list:
                if len(columns) > 1:
                    raise ValueError("Column must be a single column name")
                columns = columns[0]
        else:
            if pdt.is_integer_dtype(type(columns)) and str_only:
                if convert:
                    columns = self.get_column_names(X)[columns]
                else:
                    raise ValueError("Column must be a string, not an integer")
            if as_list and columns is not None:
                columns = [columns]
        return columns

    def append_column(self, X, name, column):
        raise NotImplementedError()

    def add_column(self, X, index, name, column):
        raise NotImplementedError()

    def rename_column(self, X, name, new_name):
        raise NotImplementedError()

    def rename_columns(self, X, mapping):
        raise NotImplementedError()

    def select_columns(
        self,
        X,
        columns=None,
        **kwargs,
    ):
        raise NotImplementedError()

    def select_column(self, X, column, **kwargs):
        raise NotImplementedError()

    def set_column(self, X, column, value):
        raise NotImplementedError()

    def set_column_names(self, X, names, new_fingerprint=None):
        raise NotImplementedError()

    def to_frame(self, X, name=None):
        raise NotImplementedError()

    def select_rows(self, X, indices, **kwargs):
        raise NotImplementedError()

    def select_row(self, X, index):
        raise NotImplementedError()

    def unique(self, X, column=None):
        raise NotImplementedError()

    def replace(self, X, column=None, mapping={}):
        raise NotImplementedError()

    def nunique(self, X, column=None):
        raise NotImplementedError()

    def argmax(self, X, axis=0):
        raise NotImplementedError()

    def ge(self, X, value=None, column=None):
        raise NotImplementedError()

    def gt(self, X, value=None, column=None):
        raise NotImplementedError()

    def le(self, X, value=None, column=None):
        raise NotImplementedError()

    def lt(self, X, value=None, column=None):
        raise NotImplementedError()

    def eq(self, X, value=None, column=None):
        raise NotImplementedError()

    def ne(self, X, value=None, column=None):
        raise NotImplementedError()

    def get_column_names(self, X, generate_cols=False):
        raise NotImplementedError()

    def get_shape(self, X):
        raise NotImplementedError()

    def iter(self, X, batch_size, drop_last_batch=False):
        raise NotImplementedError()

    def concat(self, tables, axis=0, **kwargs):
        raise NotImplementedError()

    def is_array_like(self, X, min_row=10, min_col=10):
        return False

    def get_dtypes(self, X, columns=None):
        raise NotImplementedError()

    def is_categorical(self, X, column, threshold=None):
        raise NotImplementedError()



import pandas as pd
import pandas.api.types as pdt
import pyarrow as pa

import numpy as np
from biocore.utils import logging
from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    pa_table_from_arrays_kwargs,
)

from ..base import BaseDataConverter

logger = logging.get_logger(__name__)


class NumPyConverter(BaseDataConverter):
    """
    A class that provides methods to convert data to and from NumPy arrays.
    """

    dtype = "np.ndarray"
    supports_named_columns = False

    def to_list(self, X: np.ndarray, **kwargs):
        return X.tolist()

    def to_dict(self, X: np.ndarray, names=None, **kwargs):
        if names is None:
            if len(X.shape) == 1:
                X = X.reshape(-1, 1)
            names = [f"{i}" for i in range(X.shape[1])]
        return {name: X[:, i].tolist() for i, name in enumerate(names)}

    def to_dicts(self, X: np.ndarray, names=None, **kwargs):
        if names is None:
            if len(X.shape) == 1:
                X = X.reshape(-1, 1)
            names = [f"{i}" for i in range(X.shape[1])]
        return [dict(zip(names, x)) for x in X]

    def to_numpy(self, X: np.ndarray, **kwargs):
        return X

    def to_pandas(self, X: np.ndarray, **kwargs):
        if "columns" not in kwargs:
            if X.ndim == 1:
                shape = (X.shape[0], 1)
                X = X.reshape(-1, 1)
            else:
                shape = X.shape
            kwargs["columns"] = [f"col_{i}" for i in range(shape[1])]
        return pd.DataFrame(X, **get_kwargs(kwargs, pd.DataFrame.__init__))

    def to_series(self, X: np.ndarray, **kwargs):
        if len(X.shape) == 1:
            return pd.Series(X, **get_kwargs(kwargs, pd.Series.__init__))
        else:
            raise ValueError("Cannot convert 2D array to series")

    def to_polars(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.DataFrame(X, **get_kwargs(kwargs, pl.DataFrame.__init__))

    def to_arrow(self, X: np.ndarray, **kwargs):
        if len(X.shape) == 1:
            X = X.reshape(-1, 1)
        names = kwargs.get("names", None)
        if names is None:
            names = [f"col_{i}" for i in range(X.shape[1])]
        return pa.Table.from_arrays(
            X.T, names=names, **pa_table_from_arrays_kwargs(kwargs)
        )

    def to_dataset(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset(
            self.to_arrow(X, **kwargs), **get_kwargs(kwargs, Dataset.__init__)
        )

    def to_bioset(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        return Bioset(self.to_arrow(X, **kwargs), **get_kwargs(kwargs, Bioset.__init__))

    def to_iterabledataset(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for x in self.to_dicts(X, **gen_kwargs):
                yield x

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.array as da

        return da.from_array(X, **get_kwargs(kwargs, da.from_array))

    def to_ray(self, X: np.ndarray, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_numpy(X, **get_kwargs(kwargs, ray.data.from_numpy))

    def to_csr(self, X: np.ndarray, **kwargs):
        from scipy.sparse import csr_matrix

        return csr_matrix(X)

    def _check_column(self, columns, as_list=True):
        if isinstance(columns, (list, np.ndarray)):
            if len(columns) and isinstance(columns[0], (str, np.str_)):
                raise ValueError("Column names are not supported for NumPy arrays")
            if not as_list and len(columns) > 1:
                raise ValueError("Only one column can be selected")
            columns = np.asarray(columns)
        if isinstance(columns, (str, np.str_)):
            raise ValueError("Column names are not supported for NumPy arrays")
        return columns

    def append_column(self, X: np.ndarray, name, column):
        if len(column.shape) == 1:
            column = column[:, None]
        if len(X.shape) == 1:
            return np.hstack([X[:, None], column])
        return np.hstack([X, column])

    def add_column(self, X: np.ndarray, index, name, column):
        if len(column.shape) == 1:
            column = column[:, None]
        if len(X.shape) == 1:
            X = X[:, None]
        return np.hstack([X[:, :index], column, X[:, index:]])

    def select_rows(self, X: np.ndarray, indices, **kwargs):
        return X[np.asarray(indices)]

    def select_row(self, X: np.ndarray, index):
        return X[index]

    def rename_column(self, X, name, new_name):
        logger.warning_once(
            "rename_column is not supported for list format. Returning the original data."
        )
        return X

    def rename_columns(self, X, mapping):
        logger.warning_once(
            "rename_columns is not supported for list format. Returning the original data."
        )
        return X

    def select_columns(self, X: np.ndarray, columns=None, feature_type=None, **kwargs):
        if columns is None:
            return X
        if not columns:
            return np.asarray([])

        columns = self._check_column(columns)
        if columns is None:
            return X
        return X[:, columns]

    def select_column(self, X: np.ndarray, column, **kwargs):
        column = self._check_column(column, as_list=False)
        if len(X.shape) == 1:
            return X
        return X[:, column]

    def ge(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) >= value
        return X >= value

    def gt(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) > value
        return X > value

    def le(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) <= value
        return X <= value

    def lt(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) < value
        return X < value

    def eq(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) == value
        return X == value

    def ne(self, X: np.ndarray, value, column=None):
        if column is not None:
            return self.select_column(X, column) != value
        return X != value

    def unique(self, X: np.ndarray, column=None):
        if len(X.shape) == 1:
            return np.unique(X)
        return np.unique(self.select_column(X, column))

    def nunique(self, X: np.ndarray, column=None):
        return len(self.unique(X, column))

    def replace(self, X: np.ndarray, column=None, mapping={}):
        if len(X.shape) == 1:
            return np.asarray([mapping.get(x, x) for x in X])
        return np.asarray([mapping.get(x, x) for x in self.select_column(X, column)])

    def argmax(self, X: np.ndarray, axis=0):
        if len(X.shape) == 1:
            return np.argmax(X)
        return np.argmax(X, axis=axis)

    def set_column(self, X: np.ndarray, column, value):
        if len(X.shape) == 1:
            return value
        else:
            column = self._check_column(column, as_list=False)
            X[:, column] = value
        return X

    def set_column_names(self, X: np.ndarray, names, new_fingerprint=None):
        return X

    def get_column_names(self, X: np.ndarray, generate_cols=False):
        if generate_cols:
            x_dims = X.shape
            if len(x_dims) == 1:
                return [0]
            return list(range(x_dims[1]))
        return None

    def get_column_names_by_feature_type(self, X: np.ndarray, feature_type):
        raise ValueError("NumPy does not support feature type selection")

    def get_shape(self, X: np.ndarray):
        return X.shape

    def to_frame(self, X: np.ndarray, name=None):
        if X.ndim == 1:
            return X[:, None]
        return X

    def iter(self, X: np.ndarray, batch_size, drop_last_batch=False):
        for i in range(0, len(X), batch_size):
            yield X[i : i + batch_size]

    def concat(self, tables, axis=0, **kwargs):
        return np.concatenate(tables, axis=axis)

    def is_array_like(self, X: np.ndarray, min_row=10, min_col=10):
        return True

    def get_dtypes(self, X: np.ndarray, columns=None):
        columns = self._check_column(columns)
        x_dims = X.shape
        dtype = X.dtype
        if pd.api.types.is_string_dtype(dtype):
            dtype = "string"
        elif pd.api.types.is_float_dtype(dtype):
            dtype = "float"
        if len(x_dims) == 1:
            return {0: str(dtype)}
        return {i: str(dtype) for i in range(x_dims[1])}

    def get_numeric_features(self, X: np.ndarray):
        x_dims = X.shape
        if pdt.is_numeric_dtype(X):
            if len(x_dims) == 1:
                return [0]
            return list(range(X.shape[1]))

    def is_categorical(self, X: np.ndarray, column, threshold=None):
        column = self._check_column(column, as_list=False)
        if len(X.shape) == 1:
            return False
        if pd.api.types.is_string_dtype(X) or pd.api.types.is_object_dtype(X):
            return True
        if threshold:
            n_unique = self.nunique(X, column)
            return n_unique / X.shape[0] < threshold
        return False

    def get_categorical_features(self, X: np.ndarray):
        x_dims = X.shape
        if pd.api.types.is_string_dtype(X) or pd.api.types.is_object_dtype(X):
            if len(x_dims) == 1:
                return [0]
            return list(range(X.shape[1]))



import sys
from typing import TYPE_CHECKING

import pandas as pd
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..base import BaseDataConverter

if TYPE_CHECKING:
    import dask.dataframe as dd


class DaskConverter(BaseDataConverter):
    dtype = "dd.DataFrame"

    def to_numpy(self, X: "dd.DataFrame", **kwargs):
        if "dask" in sys.modules:
            import dask.array as da

        if isinstance(X, da.Array):
            return X.compute()
        return X.compute().values

    def to_pandas(self, X: "dd.DataFrame", **kwargs):
        if "dask" in sys.modules:
            import dask.array as da

        if isinstance(X, da.Array):
            return pd.DataFrame(X.compute())
        return X.compute()

    def to_series(self, X: "dd.DataFrame", **kwargs):
        if len(X.columns) == 1:
            return X.iloc[:, 0].compute()
        if len(X.index) == 1:
            return X.compute().iloc[0, :]
        raise ValueError("Cannot convert multi-dimensional dataframe to series")

    def to_polars(self, X: "dd.DataFrame", **kwargs):
        if "polars" in sys.modules:
            import polars as pl

        return pl.from_pandas(X.compute(), **get_kwargs(kwargs, pl.from_pandas))

    def to_list(self, X: "dd.DataFrame", **kwargs):
        return X.compute().values.tolist()

    def to_dict(self, X: "dd.DataFrame", **kwargs):
        return X.compute().to_dict("list", **get_kwargs(kwargs, pd.DataFrame.to_dict))

    def to_dicts(self, X: "dd.DataFrame", **kwargs):
        return X.compute().to_dict(
            "records", **get_kwargs(kwargs, pd.DataFrame.to_dict)
        )

    def to_arrow(self, X: "dd.DataFrame", **kwargs):
        return pa.Table.from_pandas(X.compute(), **pa_table_from_pandas_kwargs(kwargs))

    def to_dataset(self, X: "dd.DataFrame", **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset.from_pandas(
            X.compute(), **get_kwargs(kwargs, Dataset.from_pandas)
        )

    def to_bioset(self, X: "dd.DataFrame", **kwargs):
        requires_backends(self.to_dataset, "bioset")
        from biosets import Bioset

        return Bioset.from_pandas(X.compute(), **get_kwargs(kwargs, Bioset.from_pandas))

    def to_iterabledataset(self, X: "dd.DataFrame", **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for _, row in X.compute().iterrows():
                yield row.to_dict()

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_dask(X, **get_kwargs(kwargs, ray.data.from_dask))



from typing import Any, Dict

import numpy as np
import pandas as pd
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    np_array_kwargs,
    np_asarray_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..dict import DictConverter


class RowDictConverter(DictConverter):
    dtype = "Dict[str, Any]"

    def to_list(self, X: Dict[str, Any], **kwargs):
        return [X]

    def to_dict(self, X: Dict[str, Any], **kwargs):
        return X

    def to_dicts(self, X: Dict[str, Any], **kwargs):
        return [X]

    def to_numpy(self, X: Dict[str, Any], **kwargs):
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            return np.asarray(self.to_list(X, **kwargs), **np_asarray_kwargs(kwargs))
        return np.array(self.to_list(X, **kwargs), **np_array_kwargs(kwargs))

    def to_pandas(self, X: Dict[str, Any], **kwargs):
        return pd.DataFrame([X], **get_kwargs(kwargs, pd.DataFrame.__init__))

    def to_series(self, X: Dict[str, Any], **kwargs):
        return pd.Series(X, **get_kwargs(kwargs, pd.Series.__init__))

    def to_polars(self, X: Dict[str, Any], **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.DataFrame([X], **get_kwargs(kwargs, pl.DataFrame.__init__))

    def to_arrow(self, X: Dict[str, Any], **kwargs):
        return pa.Table.from_pandas(
            self.to_pandas(X, **kwargs), **pa_table_from_pandas_kwargs(kwargs)
        )

    def to_dataset(self, X: Dict[str, Any], **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, Dataset.from_pandas)
        )

    def to_bioset(self, X: Dict[str, Any], **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        return Bioset.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, Bioset.from_pandas)
        )

    def to_iterabledataset(self, X: Dict[str, Any], **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            yield X

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: Dict[str, Any], **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, dd.from_pandas)
        )

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_items(X, **get_kwargs(kwargs, ray.data.from_items))



from typing import Any, Dict, List

import numpy as np
import pandas as pd
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    np_array_kwargs,
    np_asarray_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..base import BaseDataConverter


class DictsConverter(BaseDataConverter):
    dtype = "List[Dict[str, Any]]"
    supports_named_columns = True

    def to_list(self, X: List[Dict[str, Any]], **kwargs):
        return X

    def to_dict(self, X: List[Dict[str, Any]], **kwargs):
        names = kwargs.get("names", None)
        if names is None:
            if len(X) == 0:
                raise ValueError("Cannot automatically select names for empty list")
            names = list(X[0].keys())
        return {name: [x[name] for x in X] for name in names}

    def to_dicts(self, X: List[Dict[str, Any]], **kwargs):
        return X

    def to_numpy(self, X: List[Dict[str, Any]], **kwargs):
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            return np.asarray(self.to_list(X, **kwargs), **np_asarray_kwargs(kwargs))
        return np.array(self.to_list(X, **kwargs), **np_array_kwargs(kwargs))

    def to_pandas(self, X: List[Dict[str, Any]], **kwargs):
        return pd.DataFrame(X, **get_kwargs(kwargs, pd.DataFrame.__init__))

    def to_series(self, X, **kwargs):
        if len(X) == 1:
            return pd.Series(X[0], **get_kwargs(kwargs, pd.Series.__init__))
        else:
            raise ValueError("Cannot convert list of dictionaries to series")

    def to_polars(self, X: List[Dict[str, Any]], **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.from_dicts(X, **get_kwargs(kwargs, pl.from_dicts))

    def to_arrow(self, X: List[Dict[str, Any]], **kwargs):
        return pa.Table.from_pandas(
            self.to_pandas(X, **kwargs), **pa_table_from_pandas_kwargs(kwargs)
        )

    def to_dataset(self, X: List[Dict[str, Any]], **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, Dataset.from_pandas)
        )

    def to_bioset(self, X: List[Dict[str, Any]], **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        return Bioset.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, Bioset.from_pandas)
        )

    def to_iterabledataset(self, X: List[Dict[str, Any]], **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for x in X:
                yield x

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: List[Dict[str, Any]], **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        return dd.from_pandas(
            self.to_pandas(X, **kwargs), **get_kwargs(kwargs, dd.from_pandas)
        )

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_arrow(
            pa.Table.from_pylist(X), **get_kwargs(kwargs, ray.data.from_arrow)
        )

    def to_csr(self, X, **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(self.to_numpy(X), **get_kwargs(kwargs, csr_matrix.__init__))

    def append_column(self, X, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        return [{name: column[i], **X[i]} for i in range(len(X))]

    def add_column(self, X, index, name, column):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        left_cols = set(list(X[0].keys())[:index])
        right_cols = set(list(X[0].keys())[index:])
        return [
            {
                **{k: v for k, v in X[i].items() if k in left_cols},
                **{name: column[i]},
                **{k: v for k, v in X[i].items() if k in right_cols},
            }
            for i in range(len(X))
        ]

    def select_rows(self, X, indices, **kwargs):
        return [X[i] for i in indices]

    def select_row(self, X, index):
        return X[index]

    def rename_column(self, X, name, new_name):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=False)
        new_name = self._check_column(
            X, new_name, str_only=False, as_list=False, convert=False
        )
        return [{new_name if k == name else k: v for k, v in x.items()} for x in X]

    def rename_columns(self, X, mapping):
        return [{mapping.get(k, k): v for k, v in x.items()} for x in X]

    def select_columns(self, X, columns=None, feature_type=None, **kwargs):
        if columns is None:
            return X
        if not columns:
            return []

        columns = self._check_column(
            X, columns, str_only=True, as_list=True, convert=True
        )
        return [{k: x[k] for k in columns} for x in X]

    def select_column(self, X, column, **kwargs):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=True
        )
        return [x[column] for x in X]

    def unique(self, X, column=None):
        return list(set(self.select_column(X, column)))

    def replace(self, X, column=None, mapping={}):
        return self.set_column(
            X, column, [mapping.get(x[column], x[column]) for x in X]
        )

    def nunique(self, X, column=None):
        return len(self.unique(X, column))

    def argmax(self, X, axis=0):
        if axis == 0:
            return {k: np.argmax([x[k] for x in X]) for k in X[0].keys()}
        else:
            cols = []
            for x in X:
                for k, v in x.items():
                    if len(cols) <= k:
                        cols.append([])
                    cols[k].append(v)
            return {k: np.argmax(v, axis=axis) for k, v in enumerate(cols)}

    def ge(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] >= value for x in X]
        raise ValueError("A column must be specified.")

    def gt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] > value for x in X]
        raise ValueError("A column must be specified.")

    def le(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] <= value for x in X]
        raise ValueError("A column must be specified.")

    def lt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] < value for x in X]
        raise ValueError("A column must be specified.")

    def eq(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] == value for x in X]
        raise ValueError("A column must be specified.")

    def ne(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return [x[column] != value for x in X]
        raise ValueError("A column must be specified.")

    def set_column(self, X, column, value):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        for x in X:
            x[column] = value
        return X

    def set_column_names(self, X, names, new_fingerprint=None):
        names = self._check_column(X, names, str_only=True, as_list=True, convert=False)
        return [{name: x[i] for i, name in enumerate(names)} for x in X]

    def get_column_names(self, X, generate_cols=False):
        return list(X[0].keys())

    def get_shape(self, X):
        return (len(X), len(X[0]))

    def iter(self, X, batch_size, drop_last_batch=False):
        total_length = self.get_shape(X)[0]

        # Calculate the number of batches
        num_batches = total_length // batch_size
        if not drop_last_batch or total_length % batch_size == 0:
            num_batches += 1

        for batch_num in range(num_batches):
            start_idx = batch_num * batch_size
            end_idx = start_idx + batch_size
            yield X[start_idx:end_idx]

    def concat(self, tables, axis=0, **kwargs):
        if axis == 0:
            return [x for table in tables for x in table]

        out = tables[0]
        if not isinstance(out[0], (list, dict)):
            out = [[x] for x in out]

        for i, t in enumerate(tables[1:]):
            if isinstance(t, list):
                for j, x in enumerate(t):
                    if isinstance(x, dict):
                        out[j].update(x)
                    elif isinstance(x, list):
                        out[j].extend(x)
                    else:
                        out[j] += x
            else:
                for j, x in enumerate(t):
                    out[j].append(x)
        return out

    def is_array_like(self, X, min_row=10, min_col=10):
        # check first 100 rows to see if all have the same keys
        valid = True
        cols = set(X[0].keys())

        # check if the number of columns and rows are greater than the minimum required to be considered array-like
        # this is to avoid false positives for small lists of dictionaries
        if len(cols) <= min_col or len(X) <= min_row:
            return False

        for i in range(min(100, len(X))):
            if cols - set(X[i].keys()):
                valid = False
                break
        return valid

    def get_dtypes(self, X, columns=None):
        columns = self._check_column(
            X, columns, str_only=True, as_list=True, convert=True
        )
        full_name_mapper = {
            "str": "string",
            "int": "int64",
        }
        dtypes = self._check_first_n_rows(X, input_columns=columns, n=1000)
        return {
            k: full_name_mapper.get(v.__name__, v.__name__) for k, v in dtypes.items()
        }

    def is_categorical(self, X, column, threshold=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=True
        )
        if pd.api.types.is_string_dtype(self.get_dtypes(X, columns=[column])[column]):
            return True
        if threshold:
            n_unique = self.nunique(X, column)
            return n_unique / len(X) < threshold
        return False



from typing import List, Union

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from biocore.utils.import_util import (
    is_biosets_available,
    is_datasets_available,
    requires_backends,
)
from biocore.utils.inspect import (
    get_kwargs,
    pa_array_to_pandas_kwargs,
    pa_table_to_pandas_kwargs,
)

from ..base import BaseDataConverter


class ArrowConverter(BaseDataConverter):
    dtype = "pa.Table"
    supports_named_columns = True

    def to_numpy(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        if isinstance(X, (pa.ChunkedArray, pa.Array)):
            return self._arrow_array_to_numpy(X)
        return np.vstack([self._arrow_array_to_numpy(col) for col in X.columns]).T

    def to_list(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        if isinstance(X, (pa.ChunkedArray, pa.Array)):
            return self._arrow_array_to_list(X)
        return [self._arrow_array_to_list(col) for col in X.columns]

    def _arrow_array_to_list(self, pa_array: pa.Array) -> List:
        if is_datasets_available():
            from datasets.features.features import _ArrayXDExtensionType
            from datasets.formatting.formatting import (
                _is_array_with_nulls,
                _is_zero_copy_only,
            )

            if isinstance(pa_array, pa.ChunkedArray):
                if isinstance(pa_array.type, _ArrayXDExtensionType):
                    # don't call to_pylist() to preserve dtype of the fixed-size array
                    zero_copy_only = _is_zero_copy_only(
                        pa_array.type.storage_dtype, unnest=True
                    )
                    array: List = [
                        row
                        for chunk in pa_array.chunks
                        for row in chunk.to_numpy(zero_copy_only=zero_copy_only)
                    ]
                else:
                    zero_copy_only = _is_zero_copy_only(pa_array.type) and all(
                        not _is_array_with_nulls(chunk) for chunk in pa_array.chunks
                    )
                    array: List = [
                        row
                        for chunk in pa_array.chunks
                        for row in chunk.to_numpy(zero_copy_only=zero_copy_only)
                    ]
            else:
                if isinstance(pa_array.type, _ArrayXDExtensionType):
                    # don't call to_pylist() to preserve dtype of the fixed-size array
                    zero_copy_only = _is_zero_copy_only(
                        pa_array.type.storage_dtype, unnest=True
                    )
                    array: List = pa_array.to_numpy(zero_copy_only=zero_copy_only)
                else:
                    zero_copy_only = _is_zero_copy_only(
                        pa_array.type
                    ) and not _is_array_with_nulls(pa_array)
                    array: List = pa_array.to_numpy(
                        zero_copy_only=zero_copy_only
                    ).tolist()
            return array
        return pa_array.to_pylist()

    def _arrow_array_to_numpy(self, pa_array: pa.Array) -> np.ndarray:
        """This method is adapted from datasets.formatting.formatting.NumpyArrowExtractor._arrow_array_to_numpy

        Args:
            pa_array (pa.Array): _description_

        Returns:
            np.ndarray: _description_
        """
        array = self._arrow_array_to_list(pa_array)
        if len(array) > 0:
            if any(
                (
                    isinstance(x, np.ndarray)
                    and (x.dtype == object or x.shape != array[0].shape)
                )
                or (isinstance(x, float) and np.isnan(x))
                for x in array
            ):
                if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
                    return np.asarray(array, dtype=object)
                return np.array(array, copy=False, dtype=object)
        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            return np.asarray(array)
        else:
            return np.array(array, copy=False)

    def to_pandas(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        if isinstance(X, (pa.ChunkedArray, pa.Array)):
            return X.to_pandas(**pa_array_to_pandas_kwargs(kwargs))
        return X.to_pandas(**pa_table_to_pandas_kwargs(kwargs))

    def to_polars(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.from_arrow(X, **get_kwargs(kwargs, pl.from_arrow))

    def to_arrow(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        # Since X is already an Arrow table, return it as is
        return X

    def to_dict(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        if isinstance(X, (pa.ChunkedArray, pa.Array)):
            name = kwargs.get("name", "data")
            return {name: X.to_pylist()}
        return X.to_pydict()

    def to_dicts(
        self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs
    ) -> List[dict]:
        if isinstance(X, (pa.ChunkedArray, pa.Array)):
            return [{kwargs.get("name", "data"): X.to_pylist()}]

        return X.to_pylist()

    def to_dataset(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset(X, **get_kwargs(kwargs, Dataset.__init__))

    def to_bioset(self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs):
        requires_backends(self.to_bioset, "biosets")
        from biosets import Bioset

        return Bioset(X, **get_kwargs(kwargs, Bioset.__init__))

    def to_iterabledataset(
        self, X: Union[pa.Array, pa.ChunkedArray, pa.Table], **kwargs
    ):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(X):
            for row in X.to_pylist():
                yield row

        return IterableDataset.from_generator(gen)

    def to_dask(self, X, **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.dataframe as dd

        dd.from_pandas(X.to_pandas(), **get_kwargs(kwargs, dd.from_pandas))

    def to_ray(self, X, **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_arrow(X, **get_kwargs(kwargs, ray.data.from_arrow))

    def to_csr(self, X, **kwargs):
        requires_backends(self.to_csr, "scipy")
        from scipy.sparse import csr_matrix

        return csr_matrix(self.to_numpy(X), **get_kwargs(kwargs, csr_matrix.__init__))

    def append_column(self, X, name, column):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        if not isinstance(column, (pa.Array, pa.ChunkedArray)):
            if isinstance(column, pa.Table) and len(column.column_names) == 1:
                column = column.column(column.column_names[0])
            else:
                column = pa.array(column)

        return X.append_column(name, column)

    def add_column(self, X, index, name, column):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        return X.add_column(
            index, pa.field(name, pa.array(column).type), pa.array(column)
        )

    def rename_column(self, X: pa.Table, name, new_name):
        name = self._check_column(X, name, str_only=True, as_list=False, convert=False)
        new_name = self._check_column(
            X, new_name, str_only=True, as_list=False, convert=False
        )
        new_columns = [c if c != name else new_name for c in X.column_names]
        return X.rename_column(name, new_columns)

    def rename_columns(self, X, mapping):
        if isinstance(mapping, list):
            if len(mapping) == len(X.column_names):
                return X.rename_columns(mapping)
            raise ValueError(
                "Length of mapping must be equal to the number of columns in the "
                "table. Provide either a dictionary or a list of the same length as "
                "the number of columns."
            )
        new_columns = [mapping.get(c, c) for c in X.column_names]
        return X.rename_columns(new_columns)

    def select_columns(self, X, columns=None, feature_type=None, **kwargs):
        if columns is None:
            return X
        if not columns:
            return []

        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            columns = self._check_column(
                X, columns, str_only=True, as_list=False, convert=True
            )
            return pa.table({columns: X})
        columns = self._check_column(
            X, columns, str_only=True, as_list=True, convert=True
        )
        if columns:
            return X.select(columns)
        if feature_type and is_datasets_available() and is_biosets_available():
            from biosets.integration import DatasetsPatcher
            from datasets import Features

            with DatasetsPatcher():
                features = Features.from_arrow_schema(X.schema)
            return X.select(
                [f for f, v in features.items() if isinstance(v, feature_type)]
            )
        return X

    def select_column(self, X, column=None, **kwargs):
        if column is None:
            if isinstance(X, pa.Table) and X.shape[1] == 1:
                return X.column(X.column_names[0])
            return X
        column = self._check_column(
            X, column, str_only=False, as_list=False, convert=True
        )
        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            return X
        return X.column(column)

    def set_column(self, X, column, value):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        col_pos = X.column_names.index(column) if isinstance(column, str) else column
        return X.set_column(col_pos, column, value)

    def set_column_names(self, X, names, new_fingerprint=None):
        names = self._check_column(X, names, str_only=True, as_list=True, convert=False)
        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            return X
        return X.rename_columns(names)

    def select_rows(self, X, indices, **kwargs):
        if not isinstance(indices, (pa.Array, pa.ChunkedArray)):
            indices = pa.array(indices)
        return X.take(indices)

    def select_row(self, X, index):
        return X.slice(index, 1)

    def unique(self, X, column=None):
        return self.select_column(X, column).unique()

    def replace(self, X, column=None, mapping={}):
        col = self.select_column(X, column)
        new_col = pa.array([mapping.get(v.as_py(), v.as_py()) for v in col])
        return self.set_column(X, column, new_col)

    def nunique(self, X, column=None):
        return len(self.unique(X, column))

    def argmax(self, X, axis=0):
        return np.argmax(self.to_numpy(X), axis=axis)

    def ge(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.greater_equal(X, value)

    def gt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.greater(X, value, column=None)

    def le(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.less_equal(X, value, column=None)

    def lt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.less(X, value, column=None)

    def eq(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.equal(X, value, column=None)

    def ne(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        return pc.not_equal(X, value, column=None)

    def get_column_names(self, X, generate_cols=False):
        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            return ["0"]
        return X.column_names

    def get_shape(self, X):
        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            return (len(X),)
        return X.shape

    def to_frame(self, X, name=None):
        name = self._check_column(X, name, str_only=False, as_list=False, convert=True)
        if isinstance(X, (pa.Array, pa.ChunkedArray)):
            return pa.table({name: X})
        return X

    def iter(self, X, batch_size, drop_last_batch=False):
        total_length = self.get_shape(X)[0]
        # Calculate the number of batches
        num_batches = total_length // batch_size
        if not drop_last_batch or total_length % batch_size == 0:
            num_batches += 1

        for batch_num in range(num_batches):
            start_idx = batch_num * batch_size
            yield X.slice(start_idx, batch_size)

    def concat(self, tables, axis=0, **kwargs):
        if axis == 0:
            return pa.concat_tables(tables, promote="default")
        out = tables[0]
        for i, t in enumerate(tables[1:]):
            if isinstance(t, (pa.Array, pa.ChunkedArray)):
                out = self.append_column(out, str(i), t)
            else:
                for col in t.column_names:
                    out = self.append_column(out, col, t.column(col))
        return out

    def is_array_like(self, X, min_row=10, min_col=10):
        return True

    def get_dtypes(self, X, columns=None):
        dtype_map = {
            "double": "float64",
        }

        return {
            s.name: dtype_map.get(str(s.type), str(s.type))
            for s in self.select_columns(X, columns).schema
        }

    def get_numerical_features(self, X):
        dtypes = {s.name: s.type for s in X.schema}
        return [
            k
            for k, dtype in dtypes.items()
            if pa.types.is_floating(dtype) or pa.types.is_integer(dtype)
        ]

    def is_categorical(self, X, column, threshold=None):
        col = self.select_column(X, column)
        num_rows = len(col)
        if pa.types.is_dictionary(col.type) or pa.types.is_string(col.type):
            return True
        if threshold and pa.types.is_integer(col.type):
            return self.nunique(X, column) / num_rows < threshold
        return False

    def get_categorical_features(self, X):
        dtypes = {s.name: s.type for s in X.schema}
        return [
            k
            for k, dtype in dtypes.items()
            if pa.types.is_dictionary(dtype) or pa.types.is_string(dtype)
        ]



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pyarrow as pa

from biocore.utils.import_util import requires_backends
from biocore.utils.inspect import (
    get_kwargs,
    pa_table_from_pandas_kwargs,
)

from ..base import BaseDataConverter

if TYPE_CHECKING:
    import scipy.sparse as sp


class CSRConverter(BaseDataConverter):
    """
    A class that provides methods to convert data to and from CSR (Compressed Sparse Row) format.
    """

    dtype = "csr_matrix"

    def to_list(self, X: "sp.csr_matrix", **kwargs):
        return self.to_numpy(X, **kwargs).tolist()

    def to_dict(self, X: "sp.csr_matrix", names=None, **kwargs):
        if names is None:
            names = [f"{i}" for i in range(X.shape[1])]
        return {
            name: X.getcol(i).toarray().flatten().tolist()
            for i, name in enumerate(names)
        }

    def to_dicts(self, X: "sp.csr_matrix", names=None, **kwargs):
        if names is None:
            names = [f"{i}" for i in range(X.shape[1])]
        return [dict(zip(names, row.toarray().flatten())) for row in X]

    def to_numpy(self, X: "sp.csr_matrix", **kwargs):
        return X.toarray(**get_kwargs(kwargs, X.toarray))

    def to_pandas(self, X: "sp.csr_matrix", **kwargs):
        return pd.DataFrame(
            self.to_numpy(X, **kwargs), **get_kwargs(kwargs, pd.DataFrame.__init__)
        )

    def to_series(self, X: "sp.csr_matrix", **kwargs):
        if X.shape[1] == 1:
            return pd.Series(
                self.to_numpy(X, **kwargs).flatten(),
                **get_kwargs(kwargs, pd.Series.__init__),
            )
        else:
            raise ValueError("Cannot convert 2D CSR matrix to series")

    def to_polars(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_polars, "polars")
        import polars as pl

        return pl.DataFrame(
            self.to_numpy(X, **kwargs), **get_kwargs(kwargs, pl.DataFrame.__init__)
        )

    def to_arrow(self, X: "sp.csr_matrix", **kwargs):
        return pa.Table.from_pandas(
            self.to_pandas(X, **kwargs), **pa_table_from_pandas_kwargs(kwargs)
        )

    def to_dataset(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_dataset, "datasets")
        from datasets import Dataset

        return Dataset(
            self.to_arrow(X, **kwargs), **get_kwargs(kwargs, Dataset.__init__)
        )

    def to_bioset(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_dataset, "biosets")
        from biosets import Bioset

        return Bioset(self.to_arrow(X, **kwargs), **get_kwargs(kwargs, Bioset.__init__))

    def to_iterabledataset(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_iterabledataset, "datasets")
        from datasets import IterableDataset

        def gen(**gen_kwargs):
            for x in self.to_dicts(X, **gen_kwargs):
                yield x

        return IterableDataset.from_generator(
            gen, **get_kwargs(kwargs, IterableDataset.from_generator)
        )

    def to_dask(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_dask, "dask")
        import dask.array as da

        return da.from_array(X, **get_kwargs(kwargs, da.from_array))

    def to_ray(self, X: "sp.csr_matrix", **kwargs):
        requires_backends(self.to_ray, "ray")
        import ray.data

        return ray.data.from_numpy(
            self.to_numpy(X, **kwargs), **get_kwargs(kwargs, ray.data.from_numpy)
        )

    def append_column(self, X, name, column):
        requires_backends(self.append_column, "scipy")
        import scipy.sparse as sp

        if not isinstance(column, sp.csr_matrix):
            if not isinstance(column, np.ndarray):
                column = np.asarray(column)
            if column.ndim == 1:
                column = column[:, None]
            if column.shape[0] != X.shape[0]:
                raise ValueError(
                    "The new column must have the same number of rows as the input matrix."
                )

            # Convert the new column to a CSR matrix
            column = sp.csr_matrix(column)

        return sp.hstack([X, column], format="csr")

    def add_column(self, X: "sp.csr_matrix", index, name, column):
        requires_backends(self.add_column, "scipy")
        import scipy.sparse as sp

        if not isinstance(column, sp.csr_matrix):
            if not isinstance(column, np.ndarray):
                column = np.asarray(column)
            if column.ndim == 1:
                column = column[:, None]
            if column.shape[0] != X.shape[0]:
                raise ValueError(
                    "The new column must have the same number of rows as the input matrix."
                )

            # Convert the new column to a CSR matrix
            column = sp.csr_matrix(column)

        left_part = X[:, :index] if index > 0 else None
        right_part = X[:, index:] if index < X.shape[1] else None

        if left_part is not None and right_part is not None:
            X = sp.hstack([left_part, column, right_part], format="csr")
        elif left_part is not None:
            X = sp.hstack([left_part, column], format="csr")
        elif right_part is not None:
            X = sp.hstack([column, right_part], format="csr")
        else:
            X = column

        return X

    def select_rows(self, X: "sp.csr_matrix", indices, **kwargs):
        return X[indices]

    def select_row(self, X: "sp.csr_matrix", index):
        return X[index]

    def _check_column(self, columns, limit=None):
        if isinstance(columns, (list, np.ndarray)):
            if len(columns) and isinstance(columns[0], (str, np.str_)):
                raise ValueError("Column names are not supported for NumPy arrays")
            if limit and len(columns) > limit:
                raise ValueError("Too many columns selected.")
            columns = np.asarray(columns)
        if isinstance(columns, (str, np.str_)):
            raise ValueError("Column names are not supported for NumPy arrays")
        return columns

    def select_columns(self, X: "sp.csr_matrix", columns=None, **kwargs):
        requires_backends(self.select_columns, "scipy")
        import scipy.sparse as sp

        if columns is None:
            return X
        if not columns:
            return sp.csr_matrix([])
        columns = self._check_column(columns)
        if columns is None:
            return X
        return X[:, columns]

    def select_column(self, X: "sp.csr_matrix", column, **kwargs):
        column = self._check_column(column, 1)
        if isinstance(column, int):
            return X.getcol(column)
        raise ValueError("Column must be an integer index")

    def unique(self, X: "sp.csr_matrix", column=None):
        if column is not None:
            return np.unique(X.getcol(column).toarray())
        return np.unique(self.to_numpy(X))

    def nunique(self, X: "sp.csr_matrix", column=None):
        return len(self.unique(X, column))

    def replace(self, X: "sp.csr_matrix", column=None, mapping={}):
        requires_backends(self.replace, "scipy")
        import scipy.sparse as sp

        arr = self.to_numpy(X)
        if column is not None:
            col_data = arr[:, column]
            new_col_data = np.asarray([mapping.get(item, item) for item in col_data])
            arr[:, column] = new_col_data
        else:
            for i in range(arr.shape[1]):
                col_data = arr[:, i]
                new_col_data = np.asarray(
                    [mapping.get(item, item) for item in col_data]
                )
                arr[:, i] = new_col_data
        return sp.csr_matrix(arr)

    def argmax(self, X: "sp.csr_matrix", axis=0):
        if axis == 0:
            # For column-wise max, we find the argmax of each column
            # We convert to CSC.
            X_csc = X.tocsc()
            max_indices = np.zeros(X.shape[1], dtype=int)
            for col in range(X.shape[1]):
                col_array = X_csc.getcol(col).toarray().ravel()
                max_indices[col] = col_array.argmax()
            return max_indices
        elif axis == 1:
            # For row-wise max, CSR format supports efficient row operations
            max_indices = np.zeros(X.shape[0], dtype=int)
            for row in range(X.shape[0]):
                row_array = X.getrow(row).toarray().ravel()
                max_indices[row] = row_array.argmax()
            return max_indices
        else:
            raise ValueError("Axis must be 0 (columns) or 1 (rows)")

    def ge(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) >= value
        return X >= value

    def gt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) > value
        return X > value

    def le(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) <= value
        return X <= value

    def lt(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) < value
        return X < value

    def eq(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) == value
        return X == value

    def ne(self, X, value, column=None):
        column = self._check_column(
            X, column, str_only=True, as_list=False, convert=False
        )
        if column is not None:
            return X.getcol(column) != value
        return X != value

    def set_column(self, X: "sp.csr_matrix", column, value):
        requires_backends(self.set_column, "scipy")
        import scipy.sparse as sp

        if not isinstance(value, np.ndarray):
            value = np.asarray(value)
        if value.ndim == 1:
            value = value[:, None]
        if value.shape[0] != X.shape[0]:
            raise ValueError(
                "Value array must have the same number of rows as the input matrix."
            )

        # Create a new sparse column matrix
        new_col = sp.csr_matrix(value)

        # Replace the specified column with the new sparse column
        left_part = X[:, :column] if column > 0 else None
        right_part = X[:, column + 1 :] if column < X.shape[1] - 1 else None

        if left_part is not None and right_part is not None:
            X = sp.hstack([left_part, new_col, right_part], format="csr")
        elif left_part is not None:
            X = sp.hstack([left_part, new_col], format="csr")
        elif right_part is not None:
            X = sp.hstack([new_col, right_part], format="csr")
        else:
            X = new_col  # Only when X has exactly one column

        return X

    def set_column_names(self, X: "sp.csr_matrix", names, new_fingerprint=None):
        return X

    def get_column_names(self, X: np.ndarray, generate_cols=False):
        if generate_cols:
            x_dims = X.shape
            if len(x_dims) == 1:
                return [0]
            return list(range(x_dims[1]))
        return None

    def get_column_names_by_feature_type(self, X: np.ndarray, feature_type):
        raise ValueError("NumPy does not support feature type selection")

    def get_shape(self, X: "sp.csr_matrix"):
        return X.shape

    def to_frame(self, X: "sp.csr_array", **kwargs):
        requires_backends(self.to_frame, "scipy")
        import scipy.sparse as sp

        return sp.csr_matrix(X)

    def iter(self, X: "sp.csr_matrix", batch_size, drop_last_batch=False):
        for i in range(0, len(X), batch_size):
            yield X[i : i + batch_size]

    def concat(self, tables, axis=0, **kwargs):
        requires_backends(self.concat, "scipy")
        import scipy.sparse as sp

        if axis == 0:
            return sp.vstack(tables, format="csr")
        elif axis == 1:
            return sp.hstack(tables, format="csr")
        else:
            raise ValueError("Axis must be 0 (rows) or 1 (columns)")

    def is_array_like(self, X: np.ndarray, min_row=10, min_col=10):
        return True

    def get_dtypes(self, X: np.ndarray, columns=None):
        columns = self._check_column(columns)
        x_dims = X.shape
        dtype = X.dtype
        if pd.api.types.is_string_dtype(dtype):
            dtype = "string"
        if len(x_dims) == 1:
            return {0: str(dtype)}
        return {i: str(dtype) for i in range(x_dims[1])}



import inspect
import sys

import numpy as np

from biocore.utils import logging

logger = logging.get_logger(__name__)


class InvalidColumnSelectionError(Exception):
    pass


def get_kwargs(kwargs, method):
    method_args = [
        p.name
        for p in inspect.signature(method).parameters.values()
        if p != p.VAR_KEYWORD
    ]
    return {k: kwargs[k] for k in method_args if k in kwargs}


def get_required_args(method):
    """retrieves arguments with no default values"""
    return [
        p.name
        for p in inspect.signature(method).parameters.values()
        if p.default == p.empty and p != p.VAR_KEYWORD
    ]


def pa_array_from_pandas_kwargs(kwargs: dict):
    return {
        "mask": kwargs.get("mask", None),
        "type": kwargs.get("type", None),
        "safe": kwargs.get("safe", True),
        "memory_pool": kwargs.get("memory_pool", None),
    }


def pa_table_from_pandas_kwargs(kwargs: dict):
    schema = kwargs.get("schema", None)
    preserve_index = kwargs.get("preserve_index", False)
    nthreads = kwargs.get("nthreads", None)
    columns = kwargs.get("columns", None)
    safe = kwargs.get("safe", True)
    return {
        "schema": schema,
        "preserve_index": preserve_index,
        "nthreads": nthreads,
        "columns": columns,
        "safe": safe,
    }


def pa_table_to_pandas_kwargs(kwargs: dict):
    # memory_pool=None, categories=None, bool strings_to_categorical=False, bool zero_copy_only=False, bool integer_object_nulls=False, bool date_as_object=True, bool timestamp_as_object=False, bool use_threads=True, bool deduplicate_objects=True, bool ignore_metadata=False, bool safe=True, bool split_blocks=False, bool self_destruct=False, unicode maps_as_pydicts=None, types_mapper=None, bool coerce_temporal_nanoseconds=False
    memory_pool = kwargs.get("memory_pool", None)
    categories = kwargs.get("categories", None)
    strings_to_categorical = kwargs.get("strings_to_categorical", False)
    zero_copy_only = kwargs.get("zero_copy_only", False)
    integer_object_nulls = kwargs.get("integer_object_nulls", False)
    date_as_object = kwargs.get("date_as_object", True)
    timestamp_as_object = kwargs.get("timestamp_as_object", False)
    use_threads = kwargs.get("use_threads", True)
    deduplicate_objects = kwargs.get("deduplicate_objects", True)
    ignore_metadata = kwargs.get("ignore_metadata", False)
    safe = kwargs.get("safe", True)
    split_blocks = kwargs.get("split_blocks", False)
    self_destruct = kwargs.get("self_destruct", False)
    maps_as_pydicts = kwargs.get("maps_as_pydicts", None)
    types_mapper = kwargs.get("types_mapper", None)
    coerce_temporal_nanoseconds = kwargs.get("coerce_temporal_nanoseconds", False)

    return {
        "memory_pool": memory_pool,
        "categories": categories,
        "strings_to_categorical": strings_to_categorical,
        "zero_copy_only": zero_copy_only,
        "integer_object_nulls": integer_object_nulls,
        "date_as_object": date_as_object,
        "timestamp_as_object": timestamp_as_object,
        "use_threads": use_threads,
        "deduplicate_objects": deduplicate_objects,
        "ignore_metadata": ignore_metadata,
        "safe": safe,
        "split_blocks": split_blocks,
        "self_destruct": self_destruct,
        "maps_as_pydicts": maps_as_pydicts,
        "types_mapper": types_mapper,
        "coerce_temporal_nanoseconds": coerce_temporal_nanoseconds,
    }


def pa_array_to_pandas_kwargs(kwargs: dict):
    return {
        "memory_pool": kwargs.get("memory_pool", None),
        "categories": kwargs.get("categories", None),
        "strings_to_categorical": kwargs.get("strings_to_categorical", False),
        "zero_copy_only": kwargs.get("zero_copy_only", False),
        "integer_object_nulls": kwargs.get("integer_object_nulls", False),
        "date_as_object": kwargs.get("date_as_object", True),
        "timestamp_as_object": kwargs.get("timestamp_as_object", False),
        "use_threads": kwargs.get("use_threads", True),
        "deduplicate_objects": kwargs.get("deduplicate_objects", True),
        "ignore_metadata": kwargs.get("ignore_metadata", False),
        "safe": kwargs.get("safe", True),
        "split_blocks": kwargs.get("split_blocks", False),
        "self_destruct": kwargs.get("self_destruct", False),
        "maps_as_pydicts": kwargs.get("maps_as_pydicts", None),
        "types_mapper": kwargs.get("types_mapper", None),
        "coerce_temporal_nanoseconds": kwargs.get("coerce_temporal_nanoseconds", False),
    }


def pa_table_from_arrays_kwargs(kwargs: dict):
    schema = kwargs.get("schema", None)
    metadata = kwargs.get("metadata", None)
    return {"schema": schema, "metadata": metadata}


def ds_init_kwargs(kwargs: dict):
    return get_kwargs(kwargs, sys.modules["biosets"].Dataset.__init__)


def np_array_kwargs(kwargs: dict):
    return {
        "dtype": kwargs.get("dtype", None),
        "copy": kwargs.get("copy", False),
        "order": kwargs.get("order", "K"),
        "subok": kwargs.get("subok", True),
        "ndmin": kwargs.get("ndmin", 0),
        "like": kwargs.get("like", None),
    }


def np_asarray_kwargs(kwargs: dict):
    if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
        return {
            "dtype": kwargs.get("dtype", None),
            "order": kwargs.get("order", None),
            "device": kwargs.get("device", None),
            "copy": kwargs.get("copy", None),
            "like": kwargs.get("like", None),
        }
    else:
        return {
            "dtype": kwargs.get("dtype", None),
            "order": kwargs.get("order", None),
            "like": kwargs.get("like", None),
        }



from setuptools import find_packages, setup

REQUIRED_PACKAGES = [
    "numpy>=1.17",
    "pandas",
    "pyarrow>=15.0.0",
    "packaging",
]

QUALITY_REQUIRE = ["ruff>=0.1.5"]

DOCS_REQUIRE = [
    # Might need to add doc-builder and some specific deps in the future
    "s3fs",
]

TESTS_REQUIRE = [
    "pytest",
    "pytest-timeout",
    "pytest-xdist",
    "polars>=0.20.5",
    "timezones>=0.10.2",
    "biosets",
    "datasets",
]

EXTRAS_REQUIRE = {
    "polars": ["polars>=0.20.5", "timezones>=0.10.2"],
    "datasets": ["datasets"],
    "biosets": ["biosets"],
    "test": QUALITY_REQUIRE + TESTS_REQUIRE + DOCS_REQUIRE,
    "quality": QUALITY_REQUIRE,
    "docs": DOCS_REQUIRE,
}

setup(
    name="biocore",
    version="1.1.1",  # expected format is one of x.y.z.dev0, or x.y.z.rc1 or x.y.z (no to dashes, yes to dots)
    description="Bioinformatics datasets and tools for bio-family projects",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Patrick Smyth",
    author_email="psmyth1994@gmail.com",
    url="https://github.com/psmyth94/biocore",
    download_url="https://github.com/psmyth94/biocore/tags",
    license="Apache 2.0",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    python_requires=">=3.8.0,<3.12.0",
    install_requires=REQUIRED_PACKAGES,
    extras_require=EXTRAS_REQUIRE,
    zip_safe=False,  # Required for mypy to find the py.typed file
)
